#! /usr/bin/env python
# -*- coding: utf-8 -*-

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#
# = Description
#
# Analyse number and kind of reminders on users start page
#


import sys
import psycopg2

import csv
import argparse

# QUERY TEMPLATES

header=["User"]

# Number of Case
header.append("Cases")
GET_NUM_CASES = """
SELECT COUNT(*)
   FROM master_tbl_view m
   JOIN ka_status_tbl_view st ON m.id = st.master_id
   JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
  WHERE st.status = ANY (ARRAY[1, 2]) 
    AND m.phase = ANY (ARRAY[0, 1, 4, 5, 2, 3]);
"""

# = automatic_reminders
AGE_QUERY_TEMPLATE = """
    SELECT count(m.id)
      FROM master_tbl_view m
      WHERE m.bearbeiter_id = {editor}
        AND {CONDITION}
    """

# == MinAgeReminder
header.append("MinAgeReminder")
MINAGE_CONDITION = """age(m.geburtsdatum) < interval '12 years'"""

# == JustBeforeMaxAgeReminder
header.append("JustBeforeMaxAgeReminder")
JUST_BEFORE_MAXAGE_CONDITION = """
           age(m.geburtsdatum) > interval '318 months'
           AND age(m.geburtsdatum) < interval '324 months'"""

# == MaxAgeReminder
header.append("MaxAgeReminder")
MAXAGE_CONDITION = """age(m.geburtsdatum) > interval '27 years'"""

# == MaxSaveTimeReminder
header.append("MaxSavetime")
GET_MAX_SAVETIME = """
   SELECT count(m.id)
     FROM master_tbl_view m
     JOIN ka_status_tbl_view s ON s.master_id = m.id
     WHERE {CONDITION}
   """

# = Manual appointments and reminders
# == CaseAppointment
header.append("CaseAppointment")
GET_ALLCASES_APPOINTMENTLIST_SQL = """
    SELECT count(a.id) 
      FROM ka_fall_termine_tbl_view a
      JOIN master_tbl_view m ON m.id = a.master_id
      JOIN ka_status_tbl_view s ON m.id = s.master_id
      WHERE coalesce(edatum, sdatum)::date+1 >= now()::date
        AND art = 0
        AND m.bearbeiter_id = {editor:d}
        AND s.status = 1 {extrawhere}
    """

# == GlobalAppointment
header.append("GlobalAppointment")
GET_GLOBAL_APPOINTMENTLIST_SQL = """
    SELECT count(a.id)
      FROM ka_global_termine_tbl_view a
     WHERE coalesce(edatum, sdatum)::date+1 >= now()::date {extrawhere}
    """

# == CaseReminder
header.append("CaseReminder")
GET_ALLCASES_REMINDERLIST_SQL = """
    SELECT count(a.id)
      FROM ka_fall_termine_tbl_view a
      JOIN master_tbl_view m ON m.id = a.master_id
      JOIN ka_status_tbl_view s ON m.id = s.master_id
      WHERE art = 1
        AND m.bearbeiter_id = {editor:d}
        AND s.status = 1 {extrawhere}
    """



def execute(cur, query):

    try:
        cur.execute(query)
    except psycopg2.ProgrammingError as e:
        print ("ProgrammingError while executing query."
               " Details follow: {0}".format(e))

    except psycopg2.InternalError as e:
        print ("InternalError while executing query."
               " Details follow: {0}".format(e))


def trigger_condition(use_trigger):
    if use_trigger:
        return """
            (a.trigger IS NULL
             OR a.sdatum::date - a.trigger <= now()::date )"""
    return "TRUE"


def done_check_condition(use_done_check):
    if use_done_check:
        return "NOT a.erledigt"
    return "TRUE"

#
# Parse user options
#
parser = argparse.ArgumentParser(description='JMD Reminder Analysis')
parser.add_argument('-d', '--database', dest='Database', 
                    help='Database name')
parser.add_argument('-p', '--port', dest='Port', 
                    help='Port')

args = parser.parse_args()

if args.Port:
    port = args.Port
else:
    parser.print_help()
    sys.exit(1)

if args.Database:
    databases = [args.Database]
else:
    connection_global = psycopg2.connect(port=port)
    cur_global = connection_global.cursor()
    cur_global.execute("""
                   SELECT datname 
                     FROM pg_database 
                     WHERE datname SIMILAR TO 'ka\_[0-9]%' ORDER BY datname;
                     """)
    databases = [db[0] for db in cur_global]
    cur_global.close()

# Print output header
print("\t".join(header))

for database in databases:
    conn = psycopg2.connect(database=database, port=port)
    cur = conn.cursor()
    
    #
    # Get CM users
    #
    db_basename = database[:-3] # strip the last the chars ('_db')
    
    GET_CM_USER = """SELECT id, login FROM ka_benutzer_tbl 
                       WHERE pg_has_role(login, '{0}_cm'::name, 'MEMBER'::text)
                  """.format(db_basename)
    
    execute(cur, GET_CM_USER)
    
    if cur.rowcount < 1:
        continue  # no users, proceed with next db
    users = [i for i in cur]

    GET_MAX_SAVETIME_CONFIG = """SELECT wert 
                            FROM ka_konfiguration_tbl 
                            WHERE bez = 'max_speicherdauer';
                       """
    execute(cur, GET_MAX_SAVETIME_CONFIG)

    if cur.rowcount > 0:
        MaxSavetime = cur.fetchone()[0]
    else:
        MaxSavetime = None
    
    cur.close()
    
    SET_USER = """SET SESSION AUTHORIZATION '{0}'"""
    cnt=0 # replace user name with contious numbers
    for user_id, user in users:
        cur = conn.cursor()
        execute(cur, SET_USER.format(user))

        # Init user results with anonymised login name
        userResults = ["{0}_{1:04d}".format(db_basename,cnt)]
        # userResults = [user]

        # Number of Case
        execute(cur, GET_NUM_CASES)
        userResults.append("{0:d}".format(cur.fetchone()[0]))

        # = automatic_reminders
        # == MinAgeReminder
        query = AGE_QUERY_TEMPLATE.format(
                              editor=user_id, CONDITION=MINAGE_CONDITION)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))

        # == JustBeforeMaxAgeReminder
        query = AGE_QUERY_TEMPLATE.format(
                      editor=user_id, CONDITION=JUST_BEFORE_MAXAGE_CONDITION)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))

        # == MaxAgeReminder
        query = AGE_QUERY_TEMPLATE.format(
                              editor=user_id, CONDITION=MAXAGE_CONDITION)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))

        # == MaxSaveTimeReminder
        conditions = ["s.status NOT IN (3, 4, 5)"]
        if MaxSavetime:
            conditions.append(
                "extract(day from now() - s.zugriff) > {0}".format(MaxSavetime)
            )
        query = GET_MAX_SAVETIME.format(CONDITION=" AND ".join(conditions))

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))

        # = manual appointments and reminders
        extrawhere = "AND " + " AND ".join([
            trigger_condition(True),
            done_check_condition(True)])

        # == CaseAppointment
        query = GET_ALLCASES_APPOINTMENTLIST_SQL.format(
                                      editor=user_id, extrawhere=extrawhere)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))
    
        # == GlobalAppointment
        query = GET_GLOBAL_APPOINTMENTLIST_SQL.format(
                                      extrawhere=extrawhere)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))
    
        # == CaseReminder
        query = GET_ALLCASES_REMINDERLIST_SQL.format(
                                      editor=user_id, extrawhere=extrawhere)

        execute(cur, query)
        userResults.append("{0:d}".format(cur.fetchone()[0]))
    
        cur.close()
        print("\t".join(userResults))
        cnt+=1

    conn.close()

