-- Upgrade script for logbook documents
--
-- Part of mpuls/issue6856

CREATE TABLE ka_logbuch_dokumente_tbl
(
        id serial NOT NULL,
        uuid_id CHAR(36) UNIQUE,
        logbuch_id int4 REFERENCES ka_logbuch_tbl(id) ON DELETE CASCADE,
        groesse int4,
        mime varchar(256),
        daten bytea,
        name varchar(256),
        datum date,
        CONSTRAINT ka_logbuch_dokumente_tbl_pkey PRIMARY KEY (id)
);

-- Funktionen, zum Anlegen und Löschen von Logbuchanhängen
CREATE OR REPLACE FUNCTION create_logbuch_dokument_ds(lid integer, nuuid char(36), filename varchar) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
        mid integer;
        did integer;
        rid integer;
BEGIN
    SELECT INTO mid master_id FROM ka_logbuch_tbl WHERE id = lid;
    IF NOT isEditorOrStandin(mid) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;

    SELECT INTO did id from ka_logbuch_dokumente_tbl
     WHERE name = filename AND lid = logbuch_id;
    IF NOT FOUND THEN
        INSERT INTO ka_logbuch_dokumente_tbl(logbuch_id)
             VALUES (lid);
        SELECT INTO rid currval('ka_logbuch_dokumente_tbl_id_seq');
    ELSE
        RETURN did;
    END IF;

    IF nuuid IS NOT NULL THEN
        UPDATE ka_logbuch_dokumente_tbl
        SET    uuid_id = nuuid
        WHERE  id = rid;
    ELSE
        PERFORM set_uuid('ka_logbuch_dokumente_tbl', rid);
    END IF;
    RETURN rid;
END;
$$ ;

REVOKE ALL ON FUNCTION create_logbuch_dokument_ds(integer, char(36), varchar) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_logbuch_dokument_ds( integer, char(36), varchar) TO GROUP :cm_group;

CREATE OR REPLACE FUNCTION delete_logbuch_dokument_ds(did integer) RETURNS void
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    mid integer;
BEGIN
    SELECT INTO mid t.master_id
      FROM ka_logbuch_dokumente_tbl d
      JOIN ka_logbuch_tbl t ON d.logbuch_id = t.id
     WHERE d.id = did;
    IF NOT isEditorOrStandin(mid) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;
    DELETE FROM ka_logbuch_dokumente_tbl WHERE id = did;
END;
$$ ;

REVOKE ALL ON FUNCTION delete_logbuch_dokument_ds(integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_logbuch_dokument_ds(integer)
                TO GROUP :cm_group;


-- Trigger/Logging
CREATE TRIGGER t_ka_logbuch_dokumente_tbl_insert AFTER
    INSERT ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_dokumente_tbl_update AFTER
    UPDATE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_dokumente_tbl_delete AFTER
    DELETE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE OR REPLACE FUNCTION logbuch_dokumente_lastaccess_log()
RETURNS trigger AS $$
DECLARE
    mid integer;
    lid integer;
BEGIN
IF TG_OP = 'DELETE' THEN
   lid = OLD.logbuch_id;
ELSE
   lid = NEW.logbuch_id;
END IF;

SELECT INTO mid master_id
  FROM ka_logbuch_tbl_view
 WHERE id = lid;

PERFORM set_lastaccess_time(mid);

IF TG_OP = 'DELETE' THEN
   return OLD;
ELSE
   return NEW;
END IF;

END;
$$ LANGUAGE plpgsql;


CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_insert AFTER
    INSERT ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_update AFTER
    UPDATE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_delete AFTER
    DELETE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();


--
-- Views
--

CREATE OR REPLACE VIEW ka_logbuch_dokumente_tbl_view
    AS SELECT ka_logbuch_dokumente_tbl.id,
              ka_logbuch_dokumente_tbl.uuid_id,
              ka_logbuch_tbl.master_id,
              ka_logbuch_dokumente_tbl.logbuch_id,
              ka_logbuch_dokumente_tbl.groesse,
              ka_logbuch_dokumente_tbl.mime,
              ka_logbuch_dokumente_tbl.daten,
              ka_logbuch_dokumente_tbl.name,
              ka_logbuch_dokumente_tbl.datum
         FROM ka_logbuch_dokumente_tbl
         JOIN ka_logbuch_tbl
           ON ka_logbuch_dokumente_tbl.logbuch_id = ka_logbuch_tbl.id
         JOIN master_tbl ON ka_logbuch_tbl.master_id = master_tbl.id
        WHERE master_tbl.bearbeiter_id = getuserid()
           OR master_tbl.id
              IN (SELECT nm.master_id
                    FROM nm_benutzer_master_tbl nm
                   WHERE nm.benutzer_id
                         IN (SELECT nm2.gruppe_id
                               FROM nm_benutzer_gruppe_tbl nm2
                              WHERE nm2.benutzer_id = getuserid()))
           OR pg_has_role('admin','MEMBER');

CREATE OR REPLACE RULE ka_logbuch_dokumente_delete_rule
    AS ON DELETE TO ka_logbuch_dokumente_tbl_view
       DO INSTEAD NOTHING;

CREATE OR REPLACE RULE ka_logbuch_dokumente_insert_rule
    AS ON INSERT TO ka_logbuch_dokumente_tbl_view
       DO INSTEAD NOTHING;

CREATE OR REPLACE RULE ka_logbuch_dokumente_update_rule
    AS ON UPDATE TO ka_logbuch_dokumente_tbl_view
       DO INSTEAD
          UPDATE ka_logbuch_dokumente_tbl
             SET groesse = new.groesse,
                 mime = new.mime,
                 daten = new.daten,
                 name = new.name,
                 datum = new.datum
           WHERE ka_logbuch_dokumente_tbl.id = new.id;

GRANT SELECT ON ka_logbuch_dokumente_tbl_view TO GROUP :cm_group;
GRANT UPDATE ON ka_logbuch_dokumente_tbl_view TO GROUP :cm_group;
GRANT SELECT ON ka_logbuch_dokumente_tbl_view TO GROUP :admin_group;

--
-- Anonymisation
--
CREATE OR REPLACE FUNCTION anonymize_logbuch(mid integer)
RETURNS void
LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
AS $$
BEGIN
    DELETE FROM ka_logbuch_dokumente_tbl
    WHERE logbuch_id IN (SELECT id FROM ka_logbuch_tbl WHERE master_id = mid);

    -- DELETE FROM ka_logbuch_tbl WHERE master_id = mid;
    UPDATE ka_logbuch_tbl SET
        uuid_id    = NULL,
        bearbeiter = md5(lower(trim(both ' ' from coalesce(bearbeiter, '')))),
        kurz_notiz = NULL,
        notiz      = NULL
    WHERE master_id = mid;
END;
$$;

REVOKE ALL     ON FUNCTION anonymize_logbuch(integer) FROM PUBLIC;
GRANT  EXECUTE ON FUNCTION anonymize_logbuch(integer) TO GROUP :admin_group;
