-- Upgrade script for the changes to the rg_koop_* repeat groups
--
--
-- There are two changes:
--
--    - A UNIQUE constraint
--    - No instances are created when a new case is created


-- UNIQUE Constraint
--
-- This change introduces a unique constraint on the *_institution_art
-- fields in the rg_koop_* repeat groups. There may be existing
-- instances which violate that constraint, but we can easily deal with
-- that because we can simply combine multipe such instances into one
-- instance by summing the *_partner and *_schriftlich columns of all
-- groups of instances with the same master_id and *_institution_art.
--
-- For each of the repeat groups, we create a temporary table containing
-- one row for each group of instances with the same master and
-- institution_art value where each such group has more than one member,
-- because those are the groups we have to combine. Besides the
-- master_id and institution_art, each row has the minimum intance ID
-- (min_rg_id) and the sum of the partner and schriftlich columns.
--
-- We then update the rows in the original repeat group which have IDs
-- that match the min_rg_id and set their partner and schriftlich
-- columns to the appropriate sum.
--
-- Then we delete all rows from the original repeat group that
-- contributedg to the rows in the temporary table, keeping only the one
-- with min_rg_id.
--
-- The WHERE clause of the DELETE statements looks a little scary, but
-- the sub-select will be NULL and hence the comparison not true for all
-- rows of the repeat group that don't have one of the
-- master_id/institution_art combinations in the temporary table

--
-- rg_koop_1
--
CREATE TEMPORARY TABLE
    koop_1_duplicates (mid, art, min_rg_id, sum_partner, sum_schriftlich)
ON COMMIT DROP
AS (SELECT master_id, rg_koop_1_institution_art,
           MIN(id), SUM(rg_koop_1_partner), SUM(rg_koop_1_schriftlich)
      FROM rg_koop_1_tbl
     WHERE rg_koop_1_institution_art >= 0
  GROUP BY master_id, rg_koop_1_institution_art
    HAVING count(*) > 1);

UPDATE rg_koop_1_tbl rg
   SET rg_koop_1_partner = (SELECT sum_partner FROM koop_1_duplicates
                             WHERE min_rg_id = id),
       rg_koop_1_schriftlich = (SELECT sum_schriftlich FROM koop_1_duplicates
                                 WHERE min_rg_id = id)
 WHERE id IN (SELECT min_rg_id FROM koop_1_duplicates);

DELETE FROM rg_koop_1_tbl
 WHERE id <> (SELECT min_rg_id
                FROM koop_1_duplicates
               WHERE mid = master_id
                 AND rg_koop_1_institution_art = art);


--
-- rg_koop_2
--
CREATE TEMPORARY TABLE
    koop_2_duplicates (mid, art, min_rg_id, sum_partner, sum_schriftlich)
ON COMMIT DROP
AS (SELECT master_id, rg_koop_2_institution_art,
           MIN(id), SUM(rg_koop_2_partner), SUM(rg_koop_2_schriftlich)
      FROM rg_koop_2_tbl
     WHERE rg_koop_2_institution_art >= 0
  GROUP BY master_id, rg_koop_2_institution_art
    HAVING count(*) > 1);

UPDATE rg_koop_2_tbl rg
   SET rg_koop_2_partner = (SELECT sum_partner FROM koop_2_duplicates
                             WHERE min_rg_id = id),
       rg_koop_2_schriftlich = (SELECT sum_schriftlich FROM koop_2_duplicates
                                 WHERE min_rg_id = id)
 WHERE id IN (SELECT min_rg_id FROM koop_2_duplicates);

DELETE FROM rg_koop_2_tbl
 WHERE id <> (SELECT min_rg_id
                FROM koop_2_duplicates
               WHERE mid = master_id
                 AND rg_koop_2_institution_art = art);


--
-- rg_koop_3
--
CREATE TEMPORARY TABLE
    koop_3_duplicates (mid, art, min_rg_id, sum_partner, sum_schriftlich)
ON COMMIT DROP
AS (SELECT master_id, rg_koop_3_institution_art,
           MIN(id), SUM(rg_koop_3_partner), SUM(rg_koop_3_schriftlich)
      FROM rg_koop_3_tbl
     WHERE rg_koop_3_institution_art >= 0
  GROUP BY master_id, rg_koop_3_institution_art
    HAVING count(*) > 1);

UPDATE rg_koop_3_tbl rg
   SET rg_koop_3_partner = (SELECT sum_partner FROM koop_3_duplicates
                             WHERE min_rg_id = id),
       rg_koop_3_schriftlich = (SELECT sum_schriftlich FROM koop_3_duplicates
                                 WHERE min_rg_id = id)
 WHERE id IN (SELECT min_rg_id FROM koop_3_duplicates);

DELETE FROM rg_koop_3_tbl
 WHERE id <> (SELECT min_rg_id
                FROM koop_3_duplicates
               WHERE mid = master_id
                 AND rg_koop_3_institution_art = art);


--
-- rg_koop_4
--
CREATE TEMPORARY TABLE
    koop_4_duplicates (mid, art, min_rg_id, sum_partner, sum_schriftlich)
ON COMMIT DROP
AS (SELECT master_id, rg_koop_4_institution_art,
           MIN(id), SUM(rg_koop_4_partner), SUM(rg_koop_4_schriftlich)
      FROM rg_koop_4_tbl
     WHERE rg_koop_4_institution_art >= 0
  GROUP BY master_id, rg_koop_4_institution_art
    HAVING count(*) > 1);

UPDATE rg_koop_4_tbl rg
   SET rg_koop_4_partner = (SELECT sum_partner FROM koop_4_duplicates
                             WHERE min_rg_id = id),
       rg_koop_4_schriftlich = (SELECT sum_schriftlich FROM koop_4_duplicates
                                 WHERE min_rg_id = id)
 WHERE id IN (SELECT min_rg_id FROM koop_4_duplicates);

DELETE FROM rg_koop_4_tbl
 WHERE id <> (SELECT min_rg_id
                FROM koop_4_duplicates
               WHERE mid = master_id
                 AND rg_koop_4_institution_art = art);


--
-- rg_koop_5
--
CREATE TEMPORARY TABLE
    koop_5_duplicates (mid, art, min_rg_id, sum_partner, sum_schriftlich)
ON COMMIT DROP
AS (SELECT master_id, rg_koop_5_institution_art,
           MIN(id), SUM(rg_koop_5_partner), SUM(rg_koop_5_schriftlich)
      FROM rg_koop_5_tbl
     WHERE rg_koop_5_institution_art >= 0
  GROUP BY master_id, rg_koop_5_institution_art
    HAVING count(*) > 1);

UPDATE rg_koop_5_tbl rg
   SET rg_koop_5_partner = (SELECT sum_partner FROM koop_5_duplicates
                             WHERE min_rg_id = id),
       rg_koop_5_schriftlich = (SELECT sum_schriftlich FROM koop_5_duplicates
                                 WHERE min_rg_id = id)
 WHERE id IN (SELECT min_rg_id FROM koop_5_duplicates);

DELETE FROM rg_koop_5_tbl
 WHERE id <> (SELECT min_rg_id
                FROM koop_5_duplicates
               WHERE mid = master_id
                 AND rg_koop_5_institution_art = art);


--
-- Add constraints
--
CREATE UNIQUE INDEX rg_koop_1_tbl_unique_rg_koop_1_institution_art
    ON rg_koop_1_tbl (master_id, rg_koop_1_institution_art)
 WHERE (rg_koop_1_institution_art >= 0);

CREATE UNIQUE INDEX rg_koop_2_tbl_unique_rg_koop_2_institution_art
    ON rg_koop_2_tbl (master_id, rg_koop_2_institution_art)
 WHERE (rg_koop_2_institution_art >= 0);

CREATE UNIQUE INDEX rg_koop_3_tbl_unique_rg_koop_3_institution_art
    ON rg_koop_3_tbl (master_id, rg_koop_3_institution_art)
 WHERE (rg_koop_3_institution_art >= 0);

CREATE UNIQUE INDEX rg_koop_4_tbl_unique_rg_koop_4_institution_art
    ON rg_koop_4_tbl (master_id, rg_koop_4_institution_art)
 WHERE (rg_koop_4_institution_art >= 0);

CREATE UNIQUE INDEX rg_koop_5_tbl_unique_rg_koop_5_institution_art
    ON rg_koop_5_tbl (master_id, rg_koop_5_institution_art)
 WHERE (rg_koop_5_institution_art >= 0);


--
-- No instances when a new case is created
--
CREATE OR REPLACE FUNCTION create_master_tbl(nuuid char(36))
    RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER AS
$$
DECLARE
    standin integer;
    rid     integer;
BEGIN
    rid := create_master_ds(nuuid);
    IF nuuid IS NULL THEN
        PERFORM create_rg_netzwerk_tbl(rid, NULL);
      PERFORM create_rg_kjp_gruppenangebote_tbl(rid, NULL);
      PERFORM create_rg_nichtkjp_gruppenangebote_tbl(rid, NULL);
    END IF;

    RETURN rid;
END;
$$;
