# -*- coding: utf-8 -*-
#
# Copyright 2018 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

from cgi import escape

from pylons import app_globals as g

from mpulsweb.lib.base import c, render, session, abort_unless_POST

from mpulsweb.lib.helpers import url, format_object, literal
from mpulsweb.lib.helper.filters import NA
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.translation import _, ungettext
from mpulsweb.controllers.case_bundle import \
    CaseBundleController as MpulsCaseBundleController

from jmdweb.model.deletion import count_recently_deleted, record_deletion, \
    send_notification_mail


def build_case_list(cases):
    """JMD specific version of mpulsweb.controllers.case_bundle.build_case_list.

    The difference is that we hard-code the fields we use here whereas
    the other build_case_list function re-uses the list of overview
    fields. That doesn't work well in JMD because we now have more
    fields than just the name. Some of those fields are not very useful
    without an indication of what they actually are (e.g. it would now
    attempt to show the age without telling you that it's the age) or
    fields that need special treatment, like fields read from
    repeatgroups.
    """
    list = ["<ul>"]
    fields = ("name", "vorname")
    for case in cases:
        list.append("<li>%s</li>"
                    % ", ".join([escape(format_object(NA(case.get_value(field))))
                                 for field in fields]))
    list.append("</ul>")
    return list


class CaseBundleController(MpulsCaseBundleController):

    def _buildCaseList(self, case_bundle):
        case_list = case_bundle.getCases()
        return build_case_list(case_list)

    @checkRole(('cm'))
    def markDelete(self, confirmed):
        case_bundle = self._get_bundle_or_redirect()
        if len(case_bundle) > 1:
            c.dialog_title = u"Zu viele Akten zum Löschen ausgewählt!"
            c.dialog_text = literal(u"<p>Sie dürfen nur 1 Akte gleichzeitig"
                                    u" löschen.</p>"
                                    u"<p>Bitte prüfen Sie, ob Sie die Fallakten"
                                    u" tatsächlich löschen müssen:</p>"
                                    u"<p>Fallakten dürfen nur gelöscht werden,"
                                    u" wenn sie falsch oder als Testakte"
                                    u" angelegt wurden. In allen anderen Fällen"
                                    u" müssen Sie die Fallakten anonymisieren."
                                    u" Wird eine Fallakte anonymisiert,"
                                    u" werden alle personenbezogenen Daten aus"
                                    u" dieser entfernt (auch angehängte"
                                    u" Dokumente und das Tagebuch), der"
                                    u" anonymisierte Datensatz bleibt für die"
                                    u" Auswertung erhalten. Wird eine Fallakte"
                                    u" gelöscht, geht der komplette Datensatz"
                                    u" für die Auswertung verloren.</p>")
            if case_bundle.case_list:
                last_id = case_bundle.case_list[-1]
                c.url_ok = url(controller='case_overview',
                               action='index', anchor=last_id)
            else:
                c.url_ok = url(controller='case_overview', action='index')
            if session.get('casebundle'):
                del session['casebundle']
            return render('/casebundle/dialogs/failure_markdelete.mako')

        return super(CaseBundleController, self).markDelete(confirmed)

    @checkRole(('admin'))
    def delete(self, confirmed):
        confirmed = self._checkBool(confirmed)
        c.context = "../main.mako"
        case_bundle = self._get_bundle_or_redirect()
        total = case_bundle.numDatasets()

        notification_threshold = 3
        recently_deleted = count_recently_deleted()

        deletion_notice = ""
        if recently_deleted + total >= g.deletions.notification_threshold:
            deletion_notice = (u"<p>Sie haben den Schwellenwert zur Löschung"
                               u" von Fallakten erreicht. Bitte prüfen Sie,"
                               u" ob die Fallakten tatsächlich gelöscht werden"
                               u" müssen und stellen Sie diese ggf. wieder her,"
                               u" sodass der*die zuständige Bearbeiter*in die"
                               u" Fallakten anonymisieren kann.</p>"
                               u"<p><b>Hinweis: Wenn Sie diese Löschaktion"
                               u" durchführen, wird eine Nachricht an den*die"
                               u" zuständige Bundestutor*in versendet.</b></p>")

        if not confirmed:
            c.dialog_title = ungettext(u'Delete case?', u'Delete cases?',
                                       total)
            c.dialog_text = (deletion_notice
                             + (u"<p>Wollen Sie wirklich folgende Fallakte"
                                u" löschen?"
                                if total == 1 else
                                u"Wollen Sie wirklich die folgenden Fallakten"
                                u" löschen?")
                             + "<br>"
                             + "\n".join(self._buildCaseList(case_bundle))
                             + "</p>"
                             u"<p>Fallakten dürfen nur gelöscht werden, wenn"
                             u" sie falsch oder als Testakte angelegt wurden."
                             u" In allen anderen Fällen müssen Sie die"
                             u" Fallakten anonymisieren. Wird eine Fallakte"
                             u" anonymisiert, werden alle personenbezogenen"
                             u" Daten aus dieser entfernt (auch angehängte"
                             u" Dokumente und das Tagebuch),"
                             u" der anonymisierte Datensatz bleibt für die"
                             u" Auswertung erhalten. Wird eine Fallakte"
                             u" gelöscht, geht der komplette Datensatz für die"
                             u" Auswertung verloren.</p>"
                             + "<br><br>")
            c.url_yes = url(controller='case_bundle', action='delete',
                            confirmed=1)
            last_id = case_bundle.case_list[-1]
            c.url_no = url(controller='case_overview',
                           action='index', anchor=last_id)
            return render('/casemanagement/dialogs/confirm.mako')

        abort_unless_POST()
        successes = case_bundle.delete()
        failures = total - successes
        record_deletion(successes)

        mail_sent_text = ""
        new_recently_deleted = recently_deleted + successes
        if new_recently_deleted >= g.deletions.notification_threshold:
            send_notification_mail(new_recently_deleted)
            mail_sent_text = literal(u"<p>Mit dieser Löschung wurde der"
                                     u" Schwellenwert zur Löschung"
                                     u" von Fallakten überschritten"
                                     u" und eine Nachricht an den*die "
                                     u" zuständige Bundestutor*in versendet."
                                     u"</p>")

        if not failures:
            c.success_text = (
                ungettext(u"The selected case was successfully deleted.",
                          u"All selected cases were successfully deleted.",
                          total) + mail_sent_text)
            template = self.overview()
        elif successes > 0:
            # successes != total and successes > 0 means total
            # is also greater than 1, so we can always assume the
            # plural when talking about all selected cases
            c.dialog_title = _("Not all selected cases deleted")
            c.dialog_text = (
                ungettext(u"%(failures)d of the selected cases could not be"
                          u" deleted.",
                          u"%(failures)d of the selected cases could not"
                          u" be deleted.",
                          failures) % dict(failures=failures)
                + mail_sent_text
                + " " + _("Please click OK to continue."))
            template = render('/casebundle/dialogs/failure_delete_admin.mako')
        else:
            c.dialog_title = _(u"No Case deleted")
            c.dialog_text = _(u"None of the selected cases could be deleted."
                              u" Please click OK to continue.")

            template = render('/casebundle/dialogs/failure_delete_admin.mako')

        if case_bundle.case_list:
            last_id = case_bundle.case_list[-1]
            c.url_ok = url(controller='case_overview',
                           action='index', anchor=last_id)
        else:
            c.url_ok = url(controller='case_overview', action='index')

        if session.get('casebundle'):
            del session['casebundle']
        return template
