# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>

import logging

from formencode import ForEach
from formencode.validators import Int

from mpulsweb.lib.search import CaseSearch as CaseBaseSearch
from mpulsweb.lib.search import EvaluationSearch as EvaluationBaseSearch
from mpulsweb.lib.search import CaseCountSearch, \
     ExtendedCaseSearch as ExtendedCaseBaseSearch
import mpulsweb.lib.searchexpr as S

log = logging.getLogger(__name__)


class AppMixin(object):

    VALID_OPTIONS = ("gender", "type_ending")

    MULTI_OPTIONS = ("type_ending",)

    def extra_validators(self):
        validators = super(AppMixin, self).extra_validators()
        validators["type_ending"] = ForEach(Int(), convert_to_list=True)
        return validators

    def convert_form_parameters(self, options, user, extended=True):
        search =super(AppMixin, self).convert_form_parameters(options, user,
                                                        extended=extended)
        if extended:
            type_ending = options.get("type_ending")
            if type_ending:
                search["type_ending"] = list(type_ending)

        return search

    def get_gender_clause(self, options):
        return S.Field("geschlecht", options.get("gender"))

    def get_ending_clause(self, options):
        """Return the where subclause that selects based on CM ending.
        If the type_ending option was not specified, any ending will be
        OK, so this method returns 'TRUE'.  Otherwise only endings
        contained in the list value of the type_ending option will be
        selected.  If that list is empty, no ending will be OK, so no
        case will be found.
        """
        return self.generic_int_choice_clause(options, "type_ending",
                                              "art_cm_ende")

    def get_where_subclauses(self, options):
        return super(AppMixin, self).get_where_subclauses(options) + [
            self.get_ending_clause(options),
            ]


jmd_search_fields = frozenset(
    ["ARRAY(SELECT k.adresse_plz"
     "        FROM rg_kontaktdaten_tbl_view k"
     "       WHERE k.master_id = m.id)"
     " AS adresse_plz_agg",
     "ARRAY(SELECT k.mobilfunknummer"
     "        FROM rg_kontaktdaten_tbl_view k"
     "       WHERE k.master_id = m.id)"
     " AS mobilfunknummer_agg"])


class CaseSearch(AppMixin, CaseBaseSearch):

    VALID_OPTIONS = (CaseBaseSearch.VALID_OPTIONS + AppMixin.VALID_OPTIONS)

    MULTI_OPTIONS = (CaseBaseSearch.MULTI_OPTIONS + AppMixin.MULTI_OPTIONS)

    SEARCH_FIELDS = (CaseBaseSearch.SEARCH_FIELDS | jmd_search_fields)



class EvaluationSearch(AppMixin, EvaluationBaseSearch):

    VALID_OPTIONS = (EvaluationBaseSearch.VALID_OPTIONS
                     + AppMixin.VALID_OPTIONS)

    MULTI_OPTIONS = (EvaluationBaseSearch.MULTI_OPTIONS
                     + AppMixin.MULTI_OPTIONS)



class ExtendedCaseSearch(ExtendedCaseBaseSearch):

    SEARCH_FIELDS = (ExtendedCaseBaseSearch.SEARCH_FIELDS | jmd_search_fields)

    def get_order(self, options):
        sort_field = options.get('sort_field', "m.name")
        sort_order = options.get('sort_order', "desc")
        if sort_field == "name" or sort_field == "m.name":
            return "ORDER BY m.name %s, m.vorname %s" % (sort_order,
                    sort_order)
        return self.SEARCH_ORDER % (sort_field, sort_order)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
