# -*- coding: utf-8 -*-

#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
# Roland Geider <roland.geider@intevation.de>

import logging

import formencode

from pylons import request, tmpl_context as c, app_globals as g
from pylons.controllers.util import redirect

from formed.meta.semantic import ErrorItem

from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, render, session,\
    restrict_to_POST, abort_unless_POST
from mpulsweb.lib.helpers import url
from mpulsweb.lib.security import checkRole, getDbName
from mpulsweb.lib.validators import DeleteUserHelperForm, LoginCheck
from mpulsweb.model.user import UserListObject, UserObject, UserGroupList, \
    UserGroup, UserGroupFactory, can_rename_group, group_name_used
from mpulsweb.model.case import MpulsCaseOverview, MpulsCaseBundle
from mpulsweb.model.agency import MpulsAgencyOverview

from mpulsweb.lib.formedforms import FormedForm, Override, UserSetOverride


log = logging.getLogger(__name__)


def role_to_form_result(meta_node, value):
    return meta_node.get_option(value).getName()


class UserForm(FormedForm):

    def __init__(self, action, **form_parameters):
        super(UserForm, self).__init__(
            "administration_user.xml", "administration", action,
            overrides=[
                Override("login", validator=LoginCheck()),
                Override("role", to_form_result=role_to_form_result),
                ])
        self.form_parameters = form_parameters

    def create_instance(self):
        instance = super(UserForm, self).create_instance()
        for key, value in self.form_parameters.items():
            instance.set_value(key, value)
        return instance

    def from_object(self, user_object):
        instance = self.create_instance()
        instance.set_value("firstname", user_object.first_name)
        instance.set_value("lastname", user_object.last_name)
        instance.set_value("room", user_object.room)
        instance.set_value("telephone", user_object.telephone)
        instance.set_value("branchoffice", user_object.branchoffice)
        instance.set_value("login", user_object.login)

        meta_node = self.meta_structure.root.leaves["role"]
        role = user_object.getPermissionRole()
        instance.set_value("role",
                           meta_node.get_option_by_name(role).getValue())

        instance.set_value("activated", user_object.activated)
        instance.set_value("newpass", user_object.needs_passrefresh)

        return instance


class GroupForm(FormedForm):

    def __init__(self, action):
        super(GroupForm, self).__init__(
            "administration_group.xml", "administration", action,
            overrides=[UserSetOverride("members")])

    def validate_instance(self, instance, add_error):
        name = instance.get("name")
        group_id = instance.get("group_id")
        if group_id is not None:
            group_exists, rename_ok = can_rename_group(group_id, name)
            if not group_exists:
                add_error("group_id", group_id,
                          u"Die umzubenennende Gruppe existiert nicht oder"
                          u" nicht mehr!")
            if not rename_ok:
                add_error("name", name,
                          u"Es existiert bereits eine andere Gruppe mit dem"
                          u" Namen '%s'." % (name,))
        else:
            group_exists = group_name_used(name)
            if group_exists:
                add_error("name", name,
                          u'Es existiert bereits eine Benutzergruppe mit'
                          u' dem Namen "%s"!' % (name,))

        members = instance.get("members")
        if members is None:
            # members is None when no user was selected, so we actually
            # have an empty set of selected user ids.
            instance.set_value("members", set())

    def create_instance(self):
        instance = super(GroupForm, self).create_instance()
        instance.set_value("members", set())
        return instance

    def from_object(self, group):
        instance = self.create_instance()
        instance.set_value("name", group.name)
        instance.set_value("group_id", group.id)
        instance.set_value("members", group.get_member_ids())
        return instance

    def to_form_result(self, instance):
        form_result = instance.to_form_result()
        form_result["id"] = instance.get("group_id")
        form_result["members"] = instance.get("members")
        return form_result


class AdministrationController(BaseController):

    @checkRole('admin')
    def index(self):
        return self.overviewUser()

    # USER
    @checkRole('admin')
    def overviewUser(self):
        list = UserListObject()
        c.user_overview = list.getUserList()
        return render('/administration/overview_user.mako')

    _new_user_form = UserForm("new_user_action",
                              user_role_editable=True)

    @checkRole('admin')
    def newUser(self):
        instance = self._new_user_form.create_instance()
        return self._render_new_user_form(instance, {})

    def _render_new_user_form(self, instance, errors):
        c.rendered_form = self._new_user_form.render(instance, errors)
        return render('/administration/new_user.mako')

    @checkRole('admin')
    @restrict_to_POST
    def new_user_action(self):
        instance, errors = self._new_user_form.validate_request(request.params)
        if errors:
            return self._render_new_user_form(instance, errors)

        form_result = instance.to_form_result()

        c.uo = UserObject()
        c.uo.setData(form_result)
        # create user
        c.user_password = c.uo.createLogin(form_result["login"],
                                           form_result["role"],
                                           getDbName())
        self._after_new_user(instance, c.uo)
        return render('/administration/new_user_result.mako')

    def _after_new_user(self, form_result, user):
        '''Make custom code possible.'''
        pass

    @checkRole('admin')
    def showUser(self, id=None):
        id = self._checkInt(id)
        c.uo = UserObject(id)
        c.vuo = UserGroup(c.uo.standin)
        self._show_user_extra_content(c)
        # Load the agency list if application is running as evaluation server
        if g.mpuls_config.is_enabled('module', 'agency-overview'):
            agency_overview = MpulsAgencyOverview()
            agency_overview.search('sort_field:fkz;sort_order:asc')
            c.agency_list = agency_overview.getDatasets()
        return render('/administration/show_user.mako')

    def _show_user_extra_content(self, content):
        '''Make custom code possible.'''
        pass

    _edit_user_form = UserForm("edit_user_action")

    @checkRole('admin')
    def editUser(self, id=None):
        id = self._checkInt(id)
        c.uo = UserObject(id)
        instance = self._edit_user_form.from_object(c.uo)
        instance.set_value("user_id", id)
        return self._render_edit_user_form(instance, {})

    def _render_edit_user_form(self, instance, errors):
        c.rendered_form = self._edit_user_form.render(instance, errors)
        return render('/administration/edit_user.mako')

    @checkRole('admin')
    @restrict_to_POST
    def edit_user_action(self):
        instance, errors = self._edit_user_form.validate_request(request.params)
        if errors:
            return self._render_edit_user_form(instance, errors)

        form_result = instance.to_form_result()
        c.uo = UserObject(instance.get("user_id"))
        c.uo.setData(form_result)
        self._after_edit_user(form_result, c.uo)
        c.uo.store()
        c.success_text = _('User configs saved!')
        return self.overviewUser()

    def _edit_user_extra_content(self, content):
        '''Make custom code possible.'''
        pass

    def _after_edit_user(self, form_result, user):
        pass

    @checkRole('admin')
    def resetUserPassword(self, id=None, confirmed='0'):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        c.uo = UserObject(id)
        if confirmed:
            abort_unless_POST()
            c.new_password = c.uo.resetPassword()
            return render('/administration/reset_password.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = _('Reset user password?')
            c.dialog_text = "%s %s, %s" % (_('Confirm reset user password:'),
                                           c.uo.last_name, c.uo.first_name)

            c.url_yes = url(controller='administration',
                            action='resetUserPassword', id=id, confirmed=1)
            c.url_no = url(controller='administration',
                           action='showUser', id=id)
            return render('/administration/dialogs/confirm_resetpassword.mako')

    @checkRole('admin')
    def deleteUser(self, id, confirmed='0'):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        uo = UserObject(id)
        # Make sure that the user can not delete himself
        if uo.id == session.get('USER_AUTHORIZED').id:
            c.dialog_title = _('User not deleted!')
            c.dialog_text = _('Not possible to delete current login user.'
                              ' Click on okay to continue.')
            c.url_ok = url(controller='administration', action='overviewUser')
            return render('/administration/dialogs/failed_delete_user.mako')

        if confirmed:
            abort_unless_POST()
            # Check if there are cases for which the user is the editor. If
            # he is an editor then raise a dialog which allows the user to
            # transfer the cases.
            overview = MpulsCaseOverview()
            overview.search({"editor": uo.id})
            if overview.getDatasets():
                list = UserListObject()
                filter = [u.id for u in list.getAdminList()]
                filter.append(uo.id)  # the user we want to delete
                c.user_list = list.getUserList(filter)
                c.cases = overview
                log.debug(c.cases)
                c.delete_user_id = id
                c.statemachine = g.case_factory.get_state_machine()
                return render('/administration/delete_user_helper.mako')
            else:
                self._before_delete_user(id)
                uo.delete()
                c.success_for = _('User deleted!')
                c.success_text = (_('User sucessfully deleted. '))
                return self.overviewUser()

        else:
            c.delete_user_id = id
            c.context = "../main.mako"
            c.confirm_for = _('Delete user?')
            c.question = (_('Please confirm to delete user '
                            '<strong>%s, %s</strong> (login: %s).')
                          % (uo.last_name, uo.first_name, uo.login))
            c.url_yes = url(controller='administration',
                            action='deleteUser', id=id, confirmed=1)
            c.url_no = url(controller='administration',
                           action='overviewUser')
            if 'success_text' in session:
                c.success_text = session['success_text']
                del session['success_text']
                session.save()
            return render('/administration/dialogs/confirm_deleteuser.mako')

    def _before_delete_user(self, id):
        """
        Application specific actions to perform before actually deleting
        the user. For example, applications with a structure part can
        remove the rights to the user to access it.
        """
        pass

    @checkRole('admin')
    @restrict_to_POST
    def deleteUserHelper(self):
        validator = DeleteUserHelperForm()
        all_cases = MpulsCaseOverview()
        c.notification_for = _('Assign case document')
        c.form_errors = {}
        c.form_result = {}

        if 'editor' in request.params:
            try:
                form_result = validator.to_python(request.params)
                case_bundle = MpulsCaseBundle(form_result.get('case_id', []))
                new_editor = UserObject(form_result.get('editor'))
                del_user = UserObject(form_result.get('user_id'))

                all_cases.search({"editor": del_user.id})
                case_bundle.setEditor(form_result.get('editor'))
                rest = all_cases.numDatasets() - case_bundle.numDatasets()
                if rest > 0:
                    all_cases = MpulsCaseOverview()
                    all_cases.search({"editor": del_user.id})
                    c.cases = all_cases
                    c.statemachine = g.case_factory.get_state_machine()
                    list = UserListObject()
                    filter = [u.id for u in list.getAdminList()]
                    filter.append(del_user.id)  # the user we want to delete
                    c.user_list = list.getUserList(filter)
                    c.delete_user_id = del_user.id

                    if case_bundle.numDatasets() == 0:
                        c.success_text = (_(u'No case document assigned. '
                                            u'%s, %s has more case documents.')
                                          % (del_user.first_name,
                                             del_user.last_name))
                    else:
                        c.success_text = (_(u'Case document assigned to %s, %s.'
                                            u' %s, %s has more case documents.')
                                          % (new_editor.first_name,
                                             new_editor.last_name,
                                             del_user.first_name,
                                             del_user.last_name))
                    return render('/administration/delete_user_helper.mako')

                session['success_text'] = (_('Case document assigned to %s, %s.'
                                    ' The user %s, %s can now be deleted.')
                                  % (new_editor.first_name,
                                     new_editor.last_name,
                                     del_user.first_name,
                                     del_user.last_name))
                return redirect(url(controller='administration',
                                   action='deleteUser', id=del_user.id))

            except formencode.Invalid:
                return self.deleteUser(form_result.get('user_id'), '1')
        else:
            redirect(url(controller="administration", action='overviewUser'))

    # USERGROUPS
    @checkRole('admin')
    def overviewUserGroups(self):
        # Get gid of usergroups so we can filter them out.
        user = UserListObject()
        filter = [u.gid for u in user.getUserList()]

        list = UserGroupList()
        c.group_overview = list.getGroupList(filter)
        return render('/administration/overview_usergroups.mako')

    _new_group_form = GroupForm("new_group_action")

    @checkRole('admin')
    def newUserGroup(self):
        instance = self._edit_group_form.create_instance()
        return self._render_new_group_form(instance, {})

    def _render_new_group_form(self, instance, errors):
        c.rendered_form = self._new_group_form.render(instance, errors)
        return render('/administration/new_usergroup.mako')

    @checkRole('admin')
    @restrict_to_POST
    def new_group_action(self):
        instance, errors = self._new_group_form.validate_request(request.params)
        if errors:
            return self._render_new_group_form(instance, errors)

        form_result = self._new_group_form.to_form_result(instance)
        factory = UserGroupFactory()
        try:
            group = factory.create(instance.get("name"))
            group.update_members(instance.get("members"))
            group.store()
            c.success_text = _(u'User group successfully created.')
            return self.overviewUserGroups()
        except Exception, e:
            log.exception(e)
            c.dialog_title = _('Failed to create user group!')
            c.dialog_text = _('Failed to create the user group.'
                              ' Please click on OK to continue in the'
                              ' user overview.')
            c.url_ok = url(controller='administration',
                           action='overviewUserGroups')
            return render('/administration/dialogs/failed_new_usergroup.mako')

    @checkRole('admin')
    def showUserGroup(self, id=None):
        factory = UserGroupFactory()
        id = self._checkInt(id)
        group = factory.load(id)
        groupmember = group.getUser()
        c.group = group
        c.groupmember = groupmember
        return render('/administration/show_usergroup.mako')

    _edit_group_form = GroupForm("edit_group_action")

    @checkRole('admin')
    def editUserGroup(self, id=None):
        id = self._checkInt(id)
        factory = UserGroupFactory()
        group = factory.load(id)

        instance = self._edit_group_form.from_object(group)
        return self._render_edit_group_form(instance, {})

    def _render_edit_group_form(self, instance, errors):
        c.rendered_form = self._edit_group_form.render(instance, errors)
        return render('/administration/edit_usergroup.mako')

    @checkRole('admin')
    @restrict_to_POST
    def edit_group_action(self):
        instance, errors = self._edit_group_form.validate_request(request.params)
        if errors:
            return self._render_edit_group_form(instance, errors)

        form_result = self._edit_group_form.to_form_result(instance)
        factory = UserGroupFactory()
        group = factory.load(form_result["id"])
        group.setData(form_result)
        group.update_members(form_result["members"])
        group.store()
        c.group = group
        c.success_text = _('The user group was successfully edited. ')
        return self.overviewUserGroups()

    @checkRole('admin')
    def deleteUserGroup(self, id, confirmed='0'):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        factory = UserGroupFactory()
        try:
            ugo = factory.load(id)
            if confirmed:
                abort_unless_POST()
                ugo.delete()
                c.success_text = (_('The user group was successfully deleted.'))
                return self.overviewUserGroups()
            else:
                c.context = "../main.mako"
                c.dialog_title = _('Delete user group?')
                c.dialog_text = _('Do you want to delete the user group and '
                                  'lose the data?')
                c.url_yes = url(controller='administration',
                                action='deleteUserGroup', id=ugo.id,
                                confirmed=1)
                c.url_no = url(controller='administration',
                               action='overviewUserGroups')
                return render('/administration/dialogs/'
                              'confirm_deleteusergroup.mako')
        except Exception, e:
            log.exception(e)
            c.dialog_title = _('Failed to delete the user group!')
            c.dialog_text = _('Deletion of the user group failed. '
                              'Please click on OK to continue'
                              ' in the user group overview.')
            c.url_ok = url(controller='administration',
                           action='overviewUserGroups')
            return render('/administration/dialogs/failed_new_usergroup.mako')


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
