# -*- coding: utf-8 -*-

import logging

import paste

from pylons import request, tmpl_context as c, url
from pylons.controllers.util import redirect, abort
from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, render, session, \
    restrict_to_POST, abort_unless_POST

from mpulsweb.lib.security import checkRole
import mpulsweb.lib.helpers as h

from mpulsweb.model.document import CaseDocumentAPI, DocumentNotFoundException
from mpulsweb.lib.formedforms import DocumentNewForm, DocumentRenameForm


log = logging.getLogger(__name__)


casedocument_new_form = DocumentNewForm("casedocument", "new_action")


class RenameForm(DocumentRenameForm):

    def get_api(self, instance):
        return CaseDocumentAPI(instance.get("parent_id"))


casedocument_rename_form = RenameForm("casedocument", "rename_action")


class CasedocumentController(BaseController):

    @checkRole(('admin', 'cm'))
    def index(self, id):
        return self.overview(id)

    @checkRole(('admin', 'cm'))
    def overview(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        c.files = case.getDocuments()
        return render('/documents/case_overview.mako')

    def _render_new_casedocument_form(self, instance, errors):
        c.rendered_form = casedocument_new_form.render(instance, errors)
        return render('/documents/case_new.mako')

    @checkRole('cm')
    def new(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        instance = casedocument_new_form.create_instance()
        instance.set_value("id", id)
        return self._render_new_casedocument_form(instance, {})

    @checkRole('cm')
    @restrict_to_POST
    def new_action(self):
        instance, errors = casedocument_new_form.validate_request(
                                                                 request.params)

        if errors:
            return self._render_new_casedocument_form(instance, errors)

        case_id = instance.get("id")
        self._ensure_session_case(case_id)
        if not session["case"].is_editable_state():
            abort(401)

        CaseDocumentAPI(case_id).upload_documents_transaction(
                                                           instance.get("file"))

        c.success_text = (_(u"Attachment added!"))
        return self.overview(case_id)

    @checkRole(('admin', 'cm'))
    def show(self, id):
        id = self._checkInt(id)
        api = CaseDocumentAPI(None)
        try:
            document = api.load_db(id, load_contents_immediately=True)
        except DocumentNotFoundException:
            # the document could not be found, so we should send a 404
            # response to the client.
            abort(404)

        data = paste.fileapp.DataApp(document.data,
                                     content_type=document.mime,
                                     cache_control='must-revalidate',
                                     pragma='public')
        return data(request.environ, self.start_response)

    def _load_editable_case_document(self, doc_id):
        doc_id = self._checkInt(doc_id)
        api = CaseDocumentAPI(None)
        try:
            doc = api.load_db(doc_id, load_contents_immediately=False)
        except DocumentNotFoundException:
            abort(404)

        self._ensure_session_case(doc.case)
        if not session["case"].is_editable_state():
            abort(401)

        return doc

    @checkRole('cm')
    def delete(self, id, confirmed):
        doc = self._load_editable_case_document(id)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            CaseDocumentAPI(doc.case).delete_transaction(doc.id)
            c.success_for = _(u"Attachment deleted!")
            c.success_text = (_(u"Attachment deleted!"))
            return self.overview(doc.case)
        else:
            c.context = "../main.mako"
            c.confirm_for = _(u"Delete attachment?")
            c.question = _(u"Do you want to irrevocably delete the attachment"
                           u" '%s'?") % doc.name
            c.url_yes = h.url(controller='casedocument', action='delete',
                              id=id, confirmed=1)
            c.url_no = h.url(controller="casedocument", action="index",
                             id=doc.case)
            return render('/casemanagement/dialogs/'
                          'confirm_delete_attachment.mako')

    def _render_rename_form(self, instance, errors):
        c.rendered_form = casedocument_rename_form.render(instance, errors)
        return render('/casemanagement/rename_attachment.mako')


    @checkRole('cm')
    def rename(self, id):
        doc = self._load_editable_case_document(id)
        instance = casedocument_rename_form.from_object(doc)
        instance.set_value("parent_id", doc.case)
        return self._render_rename_form(instance, {})

    @checkRole('cm')
    @restrict_to_POST
    def rename_action(self):
        instance, errors = casedocument_rename_form.validate_request(
                                                                request.params)

        if errors:
            return self._render_rename_form(instance, errors)

        id = instance.get("id")

        doc = self._load_editable_case_document(id)
        CaseDocumentAPI(doc.case).rename_transaction(doc.id,
                                                     instance.get("name"))
        c.success_text = (_(u"Attachment successfully renamed."))
        return self.overview(doc.case)
