# -*- coding: utf-8 -*-
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

import logging
import datetime
from collections import defaultdict, namedtuple

from pylons import app_globals as g

from mpulsweb.lib.helpers import timedelta_in_minutes
from mpulsweb.lib.db import db

log = logging.getLogger(__name__)


# Representation of the evaluation result. The full result is a
# ResultNode with a list of ResultNode objects as children, one for each
# category, which in turn have ResultRow objects as children, one for
# each row of the evaluation result.
ResultNode = namedtuple("ResultNode", ["title", "total", "count", "children"])
ResultRow = namedtuple("ResultRow", ["title", "total", "count"])


class LogbookEvaluation:

    """Zeitaufwände aus dem Logbuch aller Fälle"""

    def __init__(self, id):
        self.id = id

        self.sql = """
            SELECT l.id, coalesce(l.dauer, '0 second'::interval) AS dauer,
                   l.art
              FROM ka_logbuch_tbl_view l
             WHERE l.master_id = %(id)s
        """

        self.sql_typ = """SELECT
            sum(coalesce(l.dauer, '0 second'::interval)) AS dauer,
            count(l.id) AS anzahl, l.typ, m.id
            FROM ka_logbuch_tbl_view l
            JOIN master_tbl_view m
            ON m.id = l.master_id
            WHERE l.master_id = %(id)s
            AND l.art != 0
            GROUP BY l.typ, m.id
            ORDER BY l.typ
        """

    def perform(self):
        categories = []
        for category in g.mpuls_config.get('logbook', 'categories'):
            categories.append((category["name"],
                               list(category.get('items', []))))
        if not categories:
            return []

        # add entry for "Keine Angabe"
        #
        # This assumes that the logbook categories as configured have
        # some suitable category at the end that is a good place to put
        # the "keine angabe" option.
        keine_angabe = "9999"
        categories[-1][1].append(keine_angabe)

        durations_by_art = defaultdict(dict)

        with db.cursor() as cur:
            cur.execute(self.sql, {'id': self.id})
            while True:
                row = cur.fetchone()
                if not row:
                    break

                logbook_id, duration, art_values = row
                duration = int(timedelta_in_minutes(duration))
                if not art_values:
                    # no art values means "Keine Angabe"
                    art_values = [keine_angabe]

                for art in art_values:
                    if art == 0:
                        continue
                    durations_by_art[str(art)][logbook_id] = duration

        descriptions = g.mpuls_config.get('logbook', 'descriptions')[0].copy()
        descriptions[keine_angabe] = "Keine Angabe"
        categorised_results = []
        all_durations = {}
        for category_name, items in categories:
            category_durations = {}
            children = []
            for item in items:
                durations = durations_by_art[item]
                category_durations.update(durations)
                children.append(ResultRow(title=descriptions[item],
                                          total=sum(durations.values()),
                                          count=len(durations)))
            all_durations.update(category_durations)

            subtotal = sum(category_durations.values())
            categorised_results.append(ResultNode(title=category_name,
                                                  total=subtotal,
                                                  count=len(category_durations),
                                                  children=children))
        return ResultNode(title="Auswertung",
                          total=sum(all_durations.values()),
                          count=len(all_durations),
                          children=categorised_results)

    def perform_type_evaluation(self):
        """ Perform an evaluation on the "typ" field of logbooks"""
        # FIXME: Remove the following line and implement this properly
        # for multiple art values (mpuls/issue7362).
        return {'total': {'amount': 0, 'time': 0, 'percent': 0}, 'types':[]}

        with db.cursor() as cur:
            cur.execute(self.sql_typ, {'id': self.id})
            rows = cur.fetchall()
            
            # Initiate statistical data. Then iterate thorugh the data to
            # generate the real ones. Bit ugly.
            result = {'total': [], 'types':[]}
            total = {'amount': 0, 'time': 0, 'percent': 0}
            
            for row in rows:
                total['amount'] = total['amount'] + row[1]
                total['time'] = total['time'] + timedelta_in_minutes(row[0])
            
            for row in rows:
                if total['time']: # don't count if no time given
                    percent = timedelta_in_minutes(row[0]) / total['time'] * 100.0
                    result['types'].append({'typ': row[2],
                                'time' : timedelta_in_minutes(row[0]),
                                'percent': percent,
                                'amount' : row[1]})
                    total['percent'] += percent
            result['total'] = total
            return result


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
