# -*- coding: utf-8 -*-
# Copyright 2012 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import functools
import datetime

import psycopg2.extras

from pylons import app_globals as g

from mpulsweb.lib.db import db
import mpulsweb.lib.searchexpr as S


SAVE_SEARCH = re.compile(r'[^\w:;\-\. ]', re.UNICODE)


log = logging.getLogger(__name__)


class AppMixin(object):

    """
        class to configure application spectific behavior of the search which
        is common to the case and evaluation search. This class can be
        overwritten in the application specific /lib/search.py
    """


class Search(object):

    VALID_OPTIONS = ("search_string",)
    MULTI_OPTIONS = ()

    SEARCH_QUERY = """
    SELECT %(fields)s
    FROM master_tbl_view m
    JOIN ka_status_tbl_view st ON m.id = st.master_id
    JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
    WHERE %(where)s
    %(order)s
    %(limit)s %(offset)s
    """

    SEARCH_FIELDS = frozenset(["m.id", "m.bearbeiter_id", "m.phase",
                               "st.status", "st.zugriff", "st.id as s_id",
                               "st.changed_editor",
                               ("ARRAY(SELECT tag_id FROM nm_tags_cases_view"
                                "      WHERE case_id = m.id) AS tags"),
                               ])

    SEARCH_ORDER = """ORDER BY %s %s"""
    GROUP_BY = """GROUP BY %s"""

    def __init__(self):
        pass

    @classmethod
    def supports_option(self, option):
        return option in self.VALID_OPTIONS

    def _parse_search_str(self, s):
        # TODO This is a very very simple approach. We should definitely
        # implement some searchengine here
        search = []
        options = {}

        if s:
            s = SAVE_SEARCH.sub(u'', s)
            option_list = [o.strip() for o in s.split(';')]
            for option in option_list:
                if option.find(":") > 0:
                    try:
                        key, value = [x.strip() for x in option.split(":")]
                        if key in self.VALID_OPTIONS:
                            if key in self.MULTI_OPTIONS:
                                options.setdefault(key, []).append(int(value))
                            else:
                                options[key] = value
                    except ValueError:
                        log.exception("Found invalid search syntax in options.")
                else:
                    search.append(option)
        options["search_string"] = search
        return options

    def check_list(self, option, values, expected_type, convert=None):
        """Check that all items in values have the type expected_type.
        The return value is a list with all items of values that have
        the expected_type.  Values of other types are omitted from the
        result, but warnings are written to the log when that happens.
        If the optional parameter convert is given, it should be a
        function or other callable object and is applied to all items in
        the return value.
        """
        checked = []
        for v in values:
            if not isinstance(v, expected_type):
                log.warning("Omitting one of the values for option %r because"
                            " it does not have type %r."
                            " Value is %r, type of value is %r",
                            option, expected_type, v, type(v))
                continue
            if convert is not None:
                v = convert(v)
            checked.append(v)
        return checked

    def check_search_options(self, options):
        """Check and sanitize the search options.

        The return value is a dictionary with the options that are
        considered valid.  With string values converted to only contain
        characters as determined by the SAVE_SEARCH regular expression.
        """
        sanitize_string = functools.partial(SAVE_SEARCH.sub, u'')
        checked = dict()
        for key, value in options.iteritems():
            if key not in self.VALID_OPTIONS:
                log.warning("Omitting option %r because the key is not valid",
                            (key, value))
                continue
            if key in self.MULTI_OPTIONS:
                value = self.check_list(key, value, int)
            elif key == "search_string":
                value = self.check_list(key, value, basestring,
                                        sanitize_string)
            elif isinstance(value, basestring):
                value = sanitize_string(value)
            elif isinstance(value, (int, datetime.date)):
                # Nothing to do for ints and dates
                pass
            else:
                log.warning("Omitting option %r because the value has the"
                            " unsupported type %r", (key, value), type(value))
                continue

            # else value is most likely an int
            checked[key] = value
        return checked

    def _get_options(self, search):
        if isinstance(search, basestring):
            options = self._parse_search_str(search)
        else:
            options = self.check_search_options(search)
        return options

    def _build_sql(self, options, sort_params={}):
        pass

    def queryDB(self, sql, parameters=None):
        """Perform the search in the DB and return the result.
        The parameters sql and parameters are passed to the cursor's
        execute method. The return vaue is a sequence of dictionaries,
        one for each row of the result set of the query..
        """
        with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
            cur.execute(sql, parameters if parameters is not None else {})
            log.debug('QueryDB: %s' % cur.query)
            return cur.fetchall()

    def perform(self, search, sort_params={}):
        """Returns the result set of a search.
        The parameter search should be a dictionary with the search
        options.  Alternatively, for backwards compatibility it may be a
        string containing the search options in the format parsed by the
        _parse_search_str method.
        """
        return self.queryDB(*self._build_sql(self._get_options(search), sort_params))

    def validate_form_params(self, params):
        return self.get_form_validator().to_python(params)

    def get_form_validator(self):
        # FIXME: We import SearchCaseForm here because importing it
        # during import doesn't work.  The validators module performs
        # translations and accesses app_globals during import, so it has
        # to be imported when a HTTP-request is being handled.  The
        # search module is imported early on when the application starts
        # so validators cannot be imported then.
        from mpulsweb.lib.validators import SearchCaseForm
        return SearchCaseForm(**self.extra_validators())

    def extra_validators(self):
        """Return additional application specific validators.
        The return value should be a dictionary mapping form parameter
        names to validators. The return value is passed as a dictionary
        with keyword arguments to the search validator.

        This default implementation returns a new empty
        dictionary. Derived classes should call this method and add
        their own validators.
        """
        return dict()

    def convert_form_parameters(self, options, user, extended):
        """Convert form parameters to search options.
        The parameter user is the user object for the current user.
        If extended is true, the search is an extended search, otherwise
        it's the more limited default search.  In the limited default
        search, some options are not included in the return value.
        """
        search_options = dict()

        # Sorting
        if options.get('sort_field'):
            search_options["sort_field"] = options.get('sort_field')
        if options.get('sort_order'):
            search_options["sort_order"] = options.get('sort_order')

        # Searchstring
        if options.get('search_str'):
            search_options["search_string"] = [options.get('search_str')]

        # Status
        state = options.get("state")
        if state:
            # for backwards compatibility we always include -1, but only
            # if the user actually wants to search based on the state,
            # i.e. when at least one state value was given.
            search_options["state"] = [-1] + list(int(s) for s in state)

        # Responsibilty
        # shown cases where the user is editor, or standin or both
        search_options["own"] = bool(options.get("own"))
        search_options["standin"] = bool(options.get("standin"))

        # Editor
        # shown cases where the user is editor or all (-1)
        if options.get("editor") and int(options.get("editor")) != -1:
            search_options["editor"] = options.get("editor")

        if extended:

            phases = self.convert_phase_form_parameters(options.get("phase",
                                                                    ()))
            if phases:
                search_options["phase"] = phases

            # Federal state
            if (options.get("federal_state") is not None):
                search_options["federal_state"] = options.get("federal_state")

            # Dates
            if (options.get("sdate") is not None
                    and options.get("edate") is not None):
                search_options["sdate"] = options.get("sdate")
                search_options["edate"] = options.get("edate")

            # gender
            if options.get("gender", -2) != -2:
                search_options["gender"] = options["gender"]

            # branch office
            if options.get("branch"):
                search_options["branch"] = options["branch"]

        return search_options

    def convert_phase_form_parameters(self, phase_options):
        """Convert phase options for self.convert_form_parameters.
        The code for this is in a separate method so that it can be
        called from derived classes explicitly. For example, in mpuls_S,
        phase search is always active, so the phase parameters must be
        converted even when extended search is not active. In JMD,
        however, this must not be done, becaus otherwise the phase
        parameters would influence even the non-extended search (which
        happens because the phase input elements are always part of the
        form, they are just rendered as 'display:none' in the case of a
        non-extended search).

        NOTE: This code may be called twice for the same search, so it
        must not have side effects!
        """
        sm = g.case_factory.get_state_machine()
        phases = []

        for pp in phase_options:
            search_ids = sm.get_state_ids_from_search_item(pp)
            phases.extend(search_ids)

        return phases

    def generic_int_choice_clause(self, options, key, column):
        """Return a search condition for an option with multiple int values.

        This method returns searchexpr.Choice instantiated with column
        as the fieldname and options.get(key) as the values to match
        agains. See Choice for details on the semantics.
        """
        return S.Choice(column, options.get(key))

    def get_allowed_viewer_clause(self, options):
        """Return the search condition that selects based on owner or standin.
        This means the search condition a case manager user can use to
        limit the search to cases that the user can see or modify due to
        the user being the cases's owner or a standin.

        The return value is searchexpr.Responsibilities instantiated
        with the values for "own" and "standin" taken from the options
        parameter with None as defaults if options does not contain
        those keys. See the Responsibilities for details on the
        semantics.

        Derived classes may override this method if necessary.
        """
        return S.Responsibilities(options.get("own"), options.get("standin"))

    def get_editor_clause(self, options):
        """Return the search condition to select cases with a specific owner.
        This is the condition an admin user can use to limit the search
        to cases of a specific case manager.

        The owner must be given by it's user id as the value of key
        "editor" in the options dictionary. The reason this is not a
        list of user ids as might be expected from the searchexpr.Editor
        class used to implement this option, is that the class follows
        the generic interface for choices in the flexible search
        implementation, whereas this method conforms the old search
        interface where the value for "editor" was a single int.

        Derived classes may override this method if necessary.
        """
        editor = options.get('editor')
        return S.Editor(editor if editor is None else [editor])

    def get_branch_clause(self, options):
        """Return the search condition to select cases based on branch office.
        Derived classes may override this method if necessary.
        """
        return S.Branch(options.get('branch'))

    def get_needle_clause(self, options):
        """Return the search condition to select cases based on search terms.
        Derived classes may override this method if necessary.
        """
        return S.Needle(["m." + f for f in self.match_fields],
                        " ".join(options.get("search_string", ())))

    def get_phases_dates_clause(self, options):
        """Return the search condition selecting cases based on phase and dates.

        The phases used for the return value are given as the value of
        the 'phase' key of the options dictionary as a list of phase
        ids.  If the options dictionary does not contain the 'phase'
        key, the value is considered to equivalent to a list containing
        all possible phase ids.

        The start and end dates are given as the values of the 'sdate'
        and 'edate' keys of the options dictionary.

        If both a start date and an end date are given, the returned
        clause matches all cases which were in one of the given phases
        during the given time.  More precisely, it matches all cases
        where the date interval of one of the phases in the phases list
        overlaps the interval specified for the search and the current
        phase is that phase or one the successor phases of that phase.

        Otherwise, the clause matches all cases that are currently in
        one of the given phases.

        Note that the phases are specified as a list.  If the options
        dictionary contains the key 'phase' but the list of phases it
        maps to is empty, the claus will be equivalent to 'FALSE' and no
        case will match.

        Derived classes may override this method if necessary.
        """
        return S.PhaseDates(options.get('phase'),
                            options.get('sdate'),
                            options.get('edate'))

    def get_phases_clause(self, options):
        """
        The phase clause depends on the acutall implemented StateMachine

        Per default the get_phase_dates_clause is used to return a sql clause
        for a PhaseStateMachine. Currently a mpuls application that doesn't use
        the default PhaseStateMachine MUST override this method to return the
        correct clause statement. In future this code should depend on the used
        StateMachine
        """
        return self.get_phases_dates_clause(options)

    def get_federalstate_clause(self, options):
        """Return the search condition to select cases based on federal state.
        Derived classes may override this method if necessary.
        """
        return S.TRUE

    def get_status_clause(self, options):
        """Return the search condition to select cases based on status.
        Derived classes may override this method if necessary.
        """
        return S.Status(options.get("state"))

    def get_gender_clause(self, options):
        """Return the search condition to select cases based on gender.
        The default implementation simply returns 'TRUE'.

        Derived classes should override this method and include 'gender'
        in their VALID_OPTIONS value.
        """
        return S.TRUE

    def get_where_subclauses(self, options):
        """Return a list with the 'where' sub-clauses.
        This base method calls the get_*_clause methods to gather the
        commonly used subclauses.  Derived classes should extend this
        method and add new clauses to the return value if they need more
        clauses.

        For a definition what a 'clause' in a returned list is, see the
        get_where method.
        """
        return [
            self.get_needle_clause(options),
            self.get_status_clause(options),
            self.get_allowed_viewer_clause(options),
            self.get_editor_clause(options),
            self.get_branch_clause(options),
            self.get_phases_clause(options),
            self.get_federalstate_clause(options),
            self.get_gender_clause(options),
        ]

    def get_where(self, options):
        """Return the 'where' clause for the search.
        The clause contains all clauses returned by self.get_where_subclauses
        combined by AND.

        Each clause must be an instance of the Expr class from the
        searchexpr module.

        The return value of this method is a triple consisting of a
        string with an SQL fragment, a dictionary with the values to use
        for the SQL fragment and a set of fields. The set of fields
        contains the fields that should be retrieved by the search
        according to the retrieve_fields method of the search conditions.
        """
        clauses = []
        for spec in self.get_where_subclauses(options):
            if not isinstance(spec, S.Expr):
                log.error("WHERE-subclause is not an Expr instance: %r", spec)
            else:
                clauses.append(spec)

        condition = S.And(clauses)
        fragment = condition.sql_expr().to_sql()
        return (fragment.sql, fragment.kw, set(condition.retrieve_fields()))

    def get_order(self, options):
        sort_field = options.get('sort_field', "m.id")
        sort_order = options.get('sort_order', "desc")
        return self.SEARCH_ORDER % (sort_field, sort_order)

    def get_limit(self, options):
        limit = options.get('limit')
        if limit is None:
            return ("", {})
        return ("LIMIT %(mpulsweb_limit)s", {"mpulsweb_limit": limit})

    def get_offset(self, options):
        offset = options.get('offset')
        if offset is None:
            return ("", {})
        return ("OFFSET %(mpulsweb_offset)s", {"mpulsweb_offset": offset})


class CaseSearch(Search):

    VALID_OPTIONS = (Search.VALID_OPTIONS
                     + ("state", "phase", "own", "standin", "editor", "branch",
                        "sdate", "edate", "sort_field", "sort_order"))
    MULTI_OPTIONS = Search.MULTI_OPTIONS + ('state', 'phase')

    def __init__(self, retrieve_fields, match_fields):
        Search.__init__(self)
        self.retrieve_fields = retrieve_fields
        self.match_fields = match_fields
        self.extra_fields = frozenset()

    def get_fields(self, options, dynamic_fields=frozenset()):
        """Return a string with the expressions for the SELECT list.
        The expressions are taken from self.SEARCH_FIELDS and
        self.retrieve_fields and the parameter dynamic_fields which
        should be a set.

        The return value is a string with the fields expressions
        separated by commas.

        As a side-effect, the fields in dynamic_fields that are neither
        in self.SEARCH_FIELDS nor in self.retrieve_fields and that refer
        to master_tbl are bound to self.extra_fields as a set. This can
        be used to determine which extra fields are already present in
        the search results and which can be shown in the search result
        without an extra DB round-trip, provided the Case class lets
        them be preset (see the Case.get_preset_fields method).
        """
        # fields predefined by self.
        static_fields = (self.SEARCH_FIELDS
                         | set("m." + field for field in self.retrieve_fields))

        # The code that shows the result only needs to handle the
        # dynamic fields that are not already covered by the static
        # fields.
        self.extra_fields = set(name[2:]
                                for name in dynamic_fields - static_fields
                                if name.startswith("m."))

        # we need to retrieve all of the static and dynamic fields
        return ", ".join(static_fields | dynamic_fields)

    def _build_sql(self, options, sort_params={}):
        sql_fields = {}
        (sql_fields['where'], where_params, fields) = self.get_where(options)
        sql_fields['order'] = self.get_order(options)
        sql_fields['order'] = self.get_order(options)
        (sql_fields['limit'], limit_params) = self.get_limit(options)
        (sql_fields['offset'], offset_params) = self.get_offset(options)
        sql_fields['fields'] = self.get_fields(options, fields)

        parameters = {}
        parameters.update(where_params)
        parameters.update(limit_params)
        parameters.update(offset_params)
        return (self.SEARCH_QUERY % sql_fields, parameters)


class ExtendedCaseSearch(Search):
    """Search which understands a S.Expr as option."""

    SEARCH_QUERY = """
    SELECT %(fields)s
    FROM master_tbl_view m
    JOIN ka_status_tbl_view st ON m.id = st.master_id
    JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
    %(joins)s
    WHERE %(where)s
    GROUP BY %(group_by)s
    %(order)s
    %(limit)s %(offset)s
    """

    JOIN_STATEMENT = """
    JOIN %(table)s ON %(on_table)s.id = %(table)s.%(on_id)s_id
    """

    def __init__(self, retrieve_fields):
        Search.__init__(self)
        self.retrieve_fields = retrieve_fields

    def get_fields(self, options, dynamic_fields=frozenset()):
        '''
        See get_fields in CaseSearch
        '''
        static_fields = (self.SEARCH_FIELDS
                         | set("m." + field for field in self.retrieve_fields))
        self.extra_fields = set(name[2:]
                                for name in dynamic_fields - static_fields
                                if name.startswith("m."))

        search_fields = set()
        for field in options.retrieve_fields():
            # Aggregate repeat group fields.
            prefix, name = field.split(".")

            if prefix == "ka_logbuch_tbl_view":
                # Logbook fields are aggregated, because the logbook
                # effectively is a repeat group. However, the field
                # 'art' must not occur in the select list because the
                # aggregation doesn't work because it's represented by
                # an array and array_agg will either work with arrays
                # (older PostgreSQL versions) and will fail in newer
                # versions because not all arrays that are accumulated
                # have the same length.
                if name != 'art':
                    search_fields.add("array_agg(%s)" % field)
                continue

            try:
                struct_path = g.structureTree.structure_path(name)
            except KeyError, e:
                # Probably a special field that is not part of the
                # structure tree. Ignore.
                continue

            if len(struct_path) > 1:
                search_fields.add("array_agg(%s)" % field)
            else:
                search_fields.add(field)

        fields_str = ", ".join(static_fields | dynamic_fields | search_fields)

        return fields_str

    def _get_options(self, search):
        return search

    def get_where(self, options):
        fragment = options.sql_expr().to_sql()
        return (fragment.sql, fragment.kw)

    def get_joins(self, options):
        joins = set()
        for field in options.retrieve_fields():
            prefix, name = field.split(".")
            if prefix in ("m", "st"):
                continue

            if prefix == "ka_logbuch_tbl_view":
                joins.add(("ka_logbuch_tbl_view", "m", "master"))
                continue

            prev = "m"
            for node in g.structureTree.structure_path(name):
                if node.getName() == name:
                    break
                tbl_name = node.getName() + "_tbl_view"
                joins.add((tbl_name, prev,
                           "master" if prev == "m" else prev))
                prev = tbl_name

        ret = ""
        for join in joins:
            table, on_table, on_id = join
            ret += self.JOIN_STATEMENT % {"table": table,
                                          "on_table": on_table,
                                          "on_id": on_id}
        return ret

    def get_group_by(self, options):
        """ Get the group by part of the sql statement.

        To properly handle repeat groups group by is used on
        for values that are not part of a repeat group.
        This might not be necessary for Postgresql > 9.1
        where a group by on m.id should be enough for all
        master table fields.
        """
        static_fields = (self.SEARCH_FIELDS
                         | set("m." + field for field in self.retrieve_fields))
        group_fields = set()
        for field in static_fields | set(options.retrieve_fields()):
            if field == "st.id as s_id":
                group_fields.add("st.id")
                continue
            if " " in field:
                continue

            prefix, name = field.split(".")
            try:
                if (prefix != "ka_logbuch_tbl_view"
                        and len(g.structureTree.structure_path(name)) < 2):
                    group_fields.add(field)
            except KeyError:
                group_fields.add(field)
        return ", ".join(group_fields)

    def _build_sql(self, options, sort_params={}):
        sql_fields = {}
        sql_fields['fields'] = self.get_fields(options)
        (sql_fields['where'], where_params) = self.get_where(options)
        sql_fields['order'] = self.get_order(sort_params)
        (sql_fields['limit'], limit_params) = "", {}
        (sql_fields['offset'], offset_params) = "", {}
        sql_fields['joins'] = self.get_joins(options)
        sql_fields['group_by'] = self.get_group_by(options)

        parameters = {}
        parameters.update(where_params)
        parameters.update(limit_params)
        parameters.update(offset_params)
        return (self.SEARCH_QUERY % sql_fields, parameters)


class CaseCountSearch(Search):

    SEARCH_QUERY = """
    SELECT COUNT(*)
    FROM master_tbl_view m
    JOIN ka_status_tbl_view st ON m.id = st.master_id
    JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
    WHERE %(where)s
    """

    VALID_OPTIONS = Search.VALID_OPTIONS
    MULTI_OPTIONS = Search.MULTI_OPTIONS + ('state', 'phase')

    def __init__(self, match_fields):
        Search.__init__(self)
        self.match_fields = match_fields

    def _build_sql(self, options, sort_params={}):
        where_clause, parameters = self.get_where(options)[:2]
        return (self.SEARCH_QUERY % dict(where=where_clause),
                parameters)


class EvaluationSearch(Search):

    SEARCH_QUERY = """
    SELECT %(fields)s
    FROM fkz_tbl_view f
    LEFT OUTER JOIN (SELECT m.fkz, m.id, st.zugriff  FROM
    master_tbl_view m
    JOIN ka_status_tbl_view st ON m.id = st.master_id
    JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
    JOIN fkz_tbl f ON f.id = m.fkz::bigint
    WHERE %(where)s) m
    ON f.id = m.fkz::bigint
    %(group)s
    %(order)s
    """

    SEARCH_FIELDS = frozenset(["f.id", "f.fkz", "f.name", "f.bundesland_id",
                               "count(m.id) as count",
                               "max(m.zugriff) as time",
                               ])

    VALID_OPTIONS = (Search.VALID_OPTIONS
                     + ("state", "phase", "own", "standin", "editor", "branch",
                        "sdate", "edate", "sort_field", "sort_order",
                        "federal_state"))
    MULTI_OPTIONS = Search.MULTI_OPTIONS + ('state', 'phase', 'federal_state')

    def __init__(self, retrieve_fields, match_fields):
        Search.__init__(self)
        self.retrieve_fields = retrieve_fields
        self.match_fields = match_fields
        self.extra_fields = frozenset()

    def get_form_validator(self):
        # FIXME: We import EvaluationSearchCaseForm here because importing it
        # during import doesn't work.  The validators module performs
        # translations and accesses app_globals during import, so it has
        # to be imported when a HTTP-request is being handled.  The
        # search module is imported early on when the application starts
        # so validators cannot be imported then.
        from mpulsweb.lib.validators import EvaluationSearchCaseForm
        return EvaluationSearchCaseForm(**self.extra_validators())

    def get_needle_clause(self, options):
        """Return the search condition to select cases based on search terms.
        Derived classes may override this method if necessary.
        """
        # FIXME: This method is almost identical to the base class
        # method. The only difference is that the match_fields are
        # fields on the table 'f' instead of 'm'. We should try to avoid
        # such repetition.
        return S.Needle(["f." + f for f in self.match_fields],
                        " ".join(options.get("search_string", ())))

    def get_federalstate_clause(self, options):
        """Return the search condition to select cases based on federal state.
        """
        return S.FederalState(options.get('federal_state'))

    def get_fields(self, options):
        return ", ".join(self.SEARCH_FIELDS)

    def get_order(self, options):
        sort_field = options.get('sort_field', "count")
        sort_order = options.get('sort_order', "desc")
        return self.SEARCH_ORDER % (sort_field, sort_order)

    def get_group(self, options):
        group_field = options.get('group_field',
                                  "f.id, f.fkz, f.name, f.bundesland_id")
        return self.GROUP_BY % (group_field)

    def _build_sql(self, options, sort_params={}):
        sql_fields = {}
        sql_fields['fields'] = self.get_fields(options)
        (sql_fields['where'], parameters) = self.get_where(options)[:2]
        sql_fields['order'] = self.get_order(options)
        sql_fields['group'] = self.get_group(options)

        SQL = self.SEARCH_QUERY % sql_fields
        return (SQL, parameters)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
