# -*- coding: utf-8 -*-
# Copyright 2014, 2015 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

"""
Code to assemble SQL expressions with parameters.
The classes and functions in this module make it relatively easy to
handle strings with SQL fragments and associated Python values that need
to be inserted into the SQL by the DB-API.
"""


def quote_for_dbapi(sql):
    """Prepare an SQL string without parameter references for use with DBAPI.
    This means that all percent characters will be doubled.
    """
    return sql.replace("%", "%%")


class SqlFragment(object):

    """A pair of an SQL string and a dictionary of values to use in the SQL.

    The string and the dictionary can be accessed with the public
    attributes sql and kw, respectively.

    The intention is that the string and dictionary can be used as the
    arguments to the DB cursor's execute method as specified in Python's
    DB-API 2.0 (PEP 249). The SQL-String should use the 'pyformat'
    paramstyle, that is, references of the form '%(name)s'. The
    dictionary should contain the value corresponding to those names.
    """

    def __init__(self, sql="", kw=None):
        self.sql = sql
        self.kw = kw if kw is not None else {}

    def __repr__(self):
        return "SqlFragment(%r, %r)" % (self.sql, self.kw)


def sql_literal(sql):
    """Create an SqlFragment from a string with SQL code and no values.
    The string will end up in the final SQL statement as is. Any
    percent-signs in the string are escaped by doubling them so that
    they're not interpreted as pyformat markers for interpolation.
    """
    return SqlFragment(quote_for_dbapi(sql))


def sql_value(key, value):
    """Creates an SqlFragment to reference a value."""
    return SqlFragment("%({0})s".format(key), {key: value})


def sql_bin_op(op, frag0, frag1):
    """Combine two SqlFragments into one with a binary infix operator."""
    return join_sql("({0} %s {1})" % op, frag0, frag1)


def join_sql(template, *args, **kw):
    """Combine multiple SqlFragments into one according to template.

    The template parameter should be a string using the format string
    syntax used by the string's format method. That syntax has the
    advantage that percent signs are not special so that it does not
    interfere with the pyformat-style interpolation of the DB API.
    """
    if args and kw:
        raise ValueError("join_sql must be called with values as either"
                         " positional arguments of keyword arguments,"
                         " not both.")
    fmt_tuple = ()
    fmt_dict = {}
    if args:
        fmt_tuple = tuple(f.sql for f in args)
        dicts = [f.kw for f in args]

    else:
        fmt_dict = dict((key, f.sql) for (key, f) in kw.iteritems())
        dicts = [f.kw for f in kw.values()]

    return SqlFragment(quote_for_dbapi(template).format(*fmt_tuple, **fmt_dict),
                       dict_union_all(dicts))



def dict_union(d1, d2):
    """Combine the dictionaries d1 and d2 into one.
    If a key is used in both dictionaries and maps to unequal values,
    the function raises a ValueError.
    """
    union = d1.copy()
    for key, value in d2.iteritems():
        if key in union:
            value1 = union[key]
            if value1 != value:
                raise ValueError("Attempting to use key %r with two"
                                 " different values: %r != %r"
                                 % (key, value1, value))
        union[key] = value
    return union


def dict_union_all(dicts):
    """Same functionality as dict_union, but for a list of dictionaries."""
    union = {}
    for d in dicts:
        union = dict_union(union, d)
    return union


class SqlExpr(object):

    """Base class for SQL-expressions using SqlFragment when creating SQL."""

    def __init__(self):
        pass

    def to_sql(self):
        """Return the expression as SqlFragment"""
        raise NotImplementedError

    def make_unique(self, ident):
        """Return a variant of ident unique enough for use in one expression.
        The return value is unique within an expression if the
        combination of ident and id(self) is unique enough. This is
        intended to guard agains accidental identifier collisions.
        """
        return ident + str(id(self))


class Field(SqlExpr):

    """A reference to a field given by its name.
    The name is directly used as an SqlFragment when creating SQL.
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Field(%r)" % (self.name,)

    def to_sql(self):
        return SqlFragment(self.name)


class Const(SqlExpr):

    """A constant python-value to interpolate into the SQL expression"""

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Const(%r)" % (self.value,)

    def to_sql(self):
        return sql_value(self.make_unique("const"), self.value)


class Literal(SqlExpr):

    """A literal SQL value."""

    def __init__(self, sql_literal):
        self.sql_literal = sql_literal

    def __repr__(self):
        return "Literal(%r)" % (self.sql_literal,)

    def to_sql(self):
        return sql_literal("(%s)" % self.sql_literal)


class Templated(SqlExpr):

    """SqlExpr defined using a template string.

    This is basically just a node in an SqlExpr built using join_sql.
    Templated accepts a template string in the same format as join_sql
    and also accepts the rest of the arguments as either positional
    arguments or keyword arguments. The value of the arguments should be
    SqlExpr instances, though. Templated converts them to SqlFragment
    instances before calling join_sql.
    """

    def __init__(self, template, *args, **kw):
        self.template = template
        self.args = args
        self.kw = kw

    def __repr__(self):
        return ("Templated(%r, %r, %r)"
                % (self.template, self.args, self.kw))

    def to_sql(self):
        return join_sql(self.template,
                        *tuple(arg.to_sql() for arg in self.args),
                        **dict((k, v.to_sql())
                               for (k, v) in self.kw.iteritems()))


class Unary(SqlExpr):

    """Base-class for unary operator invocations.

    Derived classes must overwrite the class-attribute op with a string
    containing the SQL-operator to use.
    """

    op = None

    def __init__(self, expr):
        self.expr = expr

    def __repr__(self):
        return "%s<%s>(%r)" % (self.__class__.__name__, self.op, self.expr)


class Prefix(Unary):

    """Base-class for unary prefix operators."""

    def to_sql(self):
        return join_sql("(%s {0})" % self.op, self.expr.to_sql())


class Postfix(Unary):

    """Base-class for unary postfix operators."""

    def to_sql(self):
        return join_sql("({0} %s)" % self.op, self.expr.to_sql())


class BinOp(SqlExpr):

    """Base-class for binary operator invocations.

    Derived classes must overwrite the class-attribute op with a string
    containing the SQL-operator to use.
    """

    op = None

    def __init__(self, e0, e1):
        self.e0 = e0
        self.e1 = e1

    def __repr__(self):
        return ("%s<%s>(%r, %r)"
                % (self.__class__.__name__, self.op, self.e0, self.e1))

    def to_sql(self):
        return sql_bin_op(self.op, self.e0.to_sql(), self.e1.to_sql())


class NAry(SqlExpr):

    """Base-class for n-ary operator invocations.

    Derived classes must overwrite the class-attribute op with a string
    containing the SQL-operator to use.

    In a way, this is just a special case of the repeated invocation of
    a binary operator. Given the sub-expressions e_1 .. e_n and the
    operator OP the expression built by this class is basically

      e_1 OP e_2 OP ... OP e_n
    """

    op = None

    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return ("%s<%s>(%r)"
                % (self.__class__.__name__, self.op, self.args))

    def to_sql(self):
        op = " " + self.op + " "
        placeholders = ["{%d}" % i for i in range(len(self.args))]
        template = "(" + op.join(placeholders) + ")"
        return join_sql(template, *(arg.to_sql() for arg in self.args))


class ArrayOp(SqlExpr):

    """Base-class for array operator invocations.

    Derived classes must overwrite the class-attribute op with a string
    containing the SQL-operator to use.

    This intended to be instantiated with an array subexpression, like
    Const with a Python list and used as the right sub-expression of a
    BinOp instance. For instance, the SqlExpr

        Eq(Field("status"), ArrayAny(Const([1,2,3])))

    will essentially turn into this SQL-Fragment:

        SqlFragment('(status = ANY (%(arr)s))', {'arr': [1, 2, 3]})
    """

    op = None

    def __init__(self, array):
        self.array = array

    def __repr__(self):
        return ("%s<%s>(%r)"
                % (self.__class__.__name__, self.op, self.array))

    def to_sql(self):
        return join_sql("%s ({0})" % self.op, self.array.to_sql())


class Call(SqlExpr):

    """A function call.

    Derived classes must overwrite the class-attribute op with a string
    containing the name of the function to call. The code assumes that
    the name can be used without any special processing in an SQL
    expression.
    """

    name = None

    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return "Call<%r>(%r)" % (self.name, self.args)

    def to_sql(self):
        placeholders = ["{%d}" % i for i in range(len(self.args))]
        template = "%s(%s)" % (self.name, ",".join(placeholders))
        return join_sql(template, *(arg.to_sql() for arg in self.args))


class EQ(BinOp):
    op = "="

class NE(BinOp):
    op = "<>"

class LE(BinOp):
    op = "<="

class GE(BinOp):
    op = ">="

class LT(BinOp):
    op = "<"

class GT(BinOp):
    op = ">"

class RegexMatch(BinOp):
    op = "~*"

class Overlaps(BinOp):
    """Test overlap of e.g. arrays."""
    op = "&&"


def escaped_for_like(expr):
    """Return an SqlExpr that quotes LIKE wildcards in the argument expression.
    """
    # Use PostgreSQL's escaped string literals (indicated by the
    # E-prefix) to make sure the expression works regardless of whether
    # the standard_conforming_strings setting is true or false in
    # PostgreSQL (for versions >= 9.1)
    tmpl = (r"replace(replace(replace({}, E'\\', E'\\\\'), '%', E'\\%'),"
            r" '_', E'\\_')")
    return Templated(tmpl, expr)


class Substring(BinOp):

    op = "LIKE"

    def __init__(self, e1, e2):
        pattern = NAryConcat([Literal("'%'"), escaped_for_like(e2),
                              Literal("'%'")])
        super(Substring, self).__init__(e1, pattern)


class Startwith(BinOp):

    op = "LIKE"

    def __init__(self, e1, e2):
        pattern = NAryConcat([escaped_for_like(e2), Literal("'%'")])
        super(Startwith, self).__init__(e1, pattern)


class Concat(BinOp):
    op = "||"


class NAryConcat(NAry):
    op = "||"


class NOT(Prefix):
    op = "NOT"

class IS_NULL(Postfix):
    op = "IS NULL"

class IS_NOT_NULL(Postfix):
    op = "IS NOT NULL"




TRUE = Literal("TRUE")
FALSE = Literal("FALSE")
NULL = Literal("NULL")

# PostgreSQL's now() is equivalent to the Standard CURRENT_TIMESTAMP.
CURRENT_TIMESTAMP = Literal("CURRENT_TIMESTAMP")
NOW = CURRENT_TIMESTAMP



class NAryAnd(NAry):
    op = "AND"

class NAryOr(NAry):
    op = "OR"


class ArrayAny(ArrayOp):
    op = "ANY"

class ArryAll(ArrayOp):
    op = "ALL"


class Coalesce(Call):
    name = "COALESCE"
