import os
import logging
from gettext import NullTranslations, translation

import pylons
from pylons.i18n.translation import get_lang

# Copied from pylons.i18n.translation. Adapted code to allow otional locations
# of language files

log = logging.getLogger(__name__)


def N_(value):
    """Null translation. Used to mark strings for later translation. Simply
    returns untranslated string"""
    return value


def ugettext(value):
    """Mark a string for translation. Returns the localized unicode
    string of value.

    Mark a string to be localized as follows::

        _('This should be in lots of languages')

    """
    return _ugettext(get_lang(), value)

_ = ugettext


def ungettext(singular, plural, n):
    """Mark a string for translation. Returns the localized unicode
    string of the pluralized value.

    This does a plural-forms lookup of a message id. ``singular`` is
    used as the message id for purposes of lookup in the catalog, while
    ``n`` is used to determine which plural form to use. The returned
    message is a Unicode string.

    Mark a string to be localized as follows::

        ungettext('There is %(num)d file here', 'There are %(num)d files here',
                  n) % {'num': n}

    """
    return _ungettext(get_lang(), singular, plural, n)


def _parse_list(value):
    return [val.strip() for val in value.split(",")]


def _get_dirs_and_domains():
    """
    Returns a tuple of lists or directories and domains for searching of
    translation files (.mo files)
    """
    dirs = []
    domains = []
    if pylons.config.get('mpuls.app.path.i18n'):
        # old config. keept for backwards compatibility
        domains.append(pylons.config.get('mpuls.app.instance'))
        dirs.append(pylons.config.get('mpuls.app.path.i18n'))

    if pylons.config.get('mpuls.app.i18n.domains'):
        domains.extend(_parse_list(pylons.config.get('mpuls.app.i18n.domains')))

    if pylons.config.get('mpuls.app.i18n.paths'):
        dirs.extend(_parse_list(pylons.config.get('mpuls.app.i18n.paths')))

    # Adding defaults
    conf = pylons.config.current_conf()
    dirs.append(os.path.join(conf['pylons.paths']['root'], 'i18n'))
    domains.append('mpulsweb')

    return dirs, domains


def _ugettext(lang, value):
    dirs, domains = _get_dirs_and_domains()

    for do in domains:
        for di in dirs:
            translator = _get_translator(lang, do, di)
            if translator is not None:
                translated = translator.ugettext(value)
                if translated != value:
                    # Seems to be translated
                    return translated
    return value


def _ungettext(lang, singular, plural, n):
    dirs, domains = _get_dirs_and_domains()

    for do in domains:
        for di in dirs:
            translator = _get_translator(lang, do, di)
            if translator is not None:
                translated = translator.ungettext(singular, plural, n)
                # handle 0 as plural
                if n > 1 or n == 0:
                    if translated != plural:
                        # Seems to be translated
                        return translated
                else:

                    if translated != singular:
                        # Seems to be translated
                        return translated
    return translated

translators = {}


def _get_translator(lang, domain='mpulsweb', dir=None, **kwargs):
    """Utility method to get a valid translator object from a language
    name"""
    # log.debug('getting translator for %s' % domain)
    if not lang:
        return NullTranslations()

    if not isinstance(lang, list):
        lang = [lang]

    lang=['de_DE']

    if not dir:
        conf = pylons.config.current_conf()
        dir = os.path.join(conf['pylons.paths']['root'], 'i18n')

    translator_key = "%s_%s_%s" % (domain, str(lang), dir)
    log.debug('Trying to get translator %s -----------------' % translator_key)
    if translator_key in translators:
        return translators.get(translator_key)

    try:
        translator = translation(domain, dir, languages=lang, **kwargs)
        translator.pylons_lang = lang
        translators[translator_key] = translator
        return translator
    except IOError:
        log.debug('Can not load translator for domain %s in %s' % (domain, dir))
        translators[translator_key] = None
        return None


def set_lang(lang):
    """Set the current language used for translations.

    ``lang`` should be a string or a list of strings. If a list of
    strings, the first language is set as the main and the subsequent
    languages are added as fallbacks.
    """
    translator = _get_translator(lang)
    environ = pylons.request.environ
    environ['pylons.pylons'].translator = translator
    if 'paste.registry' in environ:
        environ['paste.registry'].replace(pylons.translator, translator)
