# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <sascha.teichmann@intevation.de>
# Ludwig Reiter <ludwig.reiter@intevation.de>

import bleach
import re
import logging
from datetime import timedelta, datetime, date
import time
import xml.etree.ElementTree as ET

from HTMLParser import HTMLParser
from htmlentitydefs import name2codepoint

from pylons import app_globals as g

from openpyxl import Workbook
from openpyxl.styles import Alignment
from openpyxl.writer.excel import save_virtual_workbook

from formed.meta.data import RepeatNode, TextLeaf, IntLeaf, DateLeaf
from formed.meta.structure import StructureNode
from formed.instance.backends.xmlimport import ComponentFactory
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree

import mpulsweb.lib.helpers as h
import mpulsweb.lib.helper.filters as F
from mpulsweb.lib.db import db
from mpulsweb.lib.evaluation import LogbookEvaluation
from mpulsweb.lib.exceptions import MpulsMissingItemException
from mpulsweb.model.document import new_node, export_logbook_documents, \
     document_structure_node, import_documents_from_api, LogbookDocumentAPI


log = logging.getLogger(__name__)


ALLOWED_SORTING = ['name', 'art', 'dauer', 'datum', 'asc', 'desc', 'bearbeiter']
LOAD_LOGBOOK = """
SELECT l.id AS id, datum, bearbeiter, art, kurz_notiz, notiz, dauer, typ,
       uuid_id, master_id
FROM ka_logbuch_tbl_view l WHERE l.master_id = %%s
ORDER BY %s %s
"""

LOAD_LOGBOOK_BY_ID = """
SELECT l.id AS id, datum, bearbeiter, art, kurz_notiz, notiz, dauer, typ,
       uuid_id, master_id
FROM ka_logbuch_tbl_view l WHERE l.id = %(mid)s
"""

CREATE_LOGBOOK_ENTRY = """
SELECT create_logbuch_ds(%(mid)s, NULL)
"""

DELETE_LOGBOOK_ENTRY = """
SELECT delete_logbuch_ds(%(mid)s)
"""

UPDATE_LOGBOOK_ENTRY = """
UPDATE ka_logbuch_tbl_view SET
    datum      = %(date)s,
    art        = %(kind)s,
    kurz_notiz = %(short_notice)s,
    notiz      = %(notice)s,
    dauer      = %(duration)s,
    typ        = %(typ)s
WHERE id = %(id)s
"""


def clean_html(text):
    return bleach.clean(text, tags=['p', 'strong', 'span', 'em', 'br', 'ol', 'ul', 'li'],
                        attributes=['style'],
                        styles=['color', 'background-color',
                                'text-decoration'],
                        strip=True)


class LoadLogbookNotExistsError(MpulsMissingItemException):

    def __init__(self, value, logbook_id=None, dbname=None):
        self.value = value
        self.logbook_id = logbook_id
        self.dbname = dbname

    def __str__(self):
        return ("%s: dbname=%r, logbook_id=%r"
                % (self.value, self.dbname, self.logbook_id))


class LogbookEntry(object):

    def __init__(self):
        self.id = None
        self.datum = None
        self.kinds = []
        self.bearbeiter_name = None
        self.kurz_notiz = None
        self.notiz = None
        self.dauer = None
        self.typ = None
        self.uuid = None
        self.master_id = None

    def setByRow(self, row):
        self.id = row[0]
        self.datum = row[1]
        self.bearbeiter_name = row[2] and h.ensure_unicode(row[2]) or None
        self.kinds = row[3]
        self.kurz_notiz = row[4] and h.ensure_unicode(row[4]) or None
        self.notiz = row[5] and h.ensure_unicode(row[5]) or None
        self.dauer = row[6]
        self.typ = row[7]
        self.uuid = row[8]
        self.master_id = row[9]

    def loadById(self, id):
        self.id = id
        with db.cursor() as cur:
            cur.execute(LOAD_LOGBOOK_BY_ID, {'mid': id})
            row = cur.fetchone()
            if not row:
                raise LoadLogbookNotExistsError('logbook entry not found',
                                                logbook_id=self.id,
                                                dbname=db.database)
            self.setByRow(row)

    def setId(self, id):
        self.id = id

    def getId(self, id):
        return self.id

    def getMasterId(self):
        return self.master_id

    def getDate(self):
        return h.dd_mm_YYYY(self.datum)

    def getTime(self):
        return h.HH_MM(self.datum)

    def setDate(self, datum):
        self.datum = datum

    def getEditor(self):
        return self.bearbeiter_name

    def getKindsAsText(self, empty='-/-'):
        if not self.kinds:
            return empty

        descriptions = g.mpuls_config.get('logbook', 'descriptions')[0]
        texts = []
        for kind in self.kinds:
            if kind == -2:
                text = "%s" % self.getShortNotice(empty)
            else:
                text = descriptions.get(str(kind), empty)
            texts.append(text)
        return ", ".join(texts)

    def setKind(self, kind):
        """Set the entry kind to the single given value."""
        self.kinds = [kind]

    def get_kinds(self):
        return self.kinds

    def set_kinds(self, kinds):
        self.kinds = kinds

    def setTyp(self, typ):
        self.typ = typ

    def getTyp(self):
        return self.typ

    def getShortNotice(self, empty='-/-'):
        if self.kurz_notiz is not None:
            return self.kurz_notiz
        return empty

    def setShortNotice(self, kurz_notiz):
        self.kurz_notiz = kurz_notiz

    def getKindAsInt(self):
        """Return the first of the kinds as an int. None if no kind is set.

        This method is only present for temporary backwards
        compatibility while support for multiple kinds is implemented.
        """
        if self.kinds:
            return self.kinds[0]
        return None

    def isSelectedKind(self, kind):
        """Return whether the kind is the selected kind.

        This method only looks at the first of the entry's kinds because
        this method is only present for temporary backwards
        compatibility while support for multiple kinds is implemented.
        """
        return kind == self.getKindAsInt() and "selected" or ""

    def setNotice(self, notice):
        self.notiz = notice

    def getNotice(self, empty=''):
        if self.notiz:
            return self.notiz
        return empty

    def getDuration(self, empty=''):
        if self.dauer is None:
            return empty
        return "%d" % h.timedelta_in_minutes(self.dauer)

    def setDuration(self, duration):
        self.dauer = duration

    def is_editable_in_case(self, case):
        """Return whether self can be edited as part of the given case object.
        The case object must represent the case to which the entry
        belongs. It's passed as an argument and not loaded automatically
        because in most cases a suitable case object is already
        available and thus we can avoid extra database round-trips.
        """
        assert self.master_id == case.id
        return self.getKindAsInt() != 0 and case.is_editable_state()

    def store(self, cur, mid, uid):
        if self.id is None:
            cur.execute(CREATE_LOGBOOK_ENTRY, {"mid": mid})
            row = cur.fetchone()
            if not row:
                raise StandardError("Cannot create log book entry")
            self.id = row[0]

        cur.execute(UPDATE_LOGBOOK_ENTRY,
                    {"date": self.datum,
                     "kind": self.kinds,
                     "short_notice": self.kurz_notiz,
                     "notice": self.notiz,
                     "duration": self.dauer,
                     "typ": self.typ,
                     "id": self.id})

    def storeForUser(self, uid):
        db.execute(UPDATE_LOGBOOK_ENTRY,
                   {"date": self.datum,
                    "kind": self.kinds,
                    "short_notice": self.kurz_notiz,
                    "notice": self.notiz,
                    "duration": self.dauer,
                    "typ": self.typ,
                    "id": self.id})

    def get_data_for_xlsx(self):
        """Return data used for the logbook xlsx export,"""
        def empty_if_none(value):
            if value is None:
                return ""
            return value

        # FIXME: Using getKindsAsText may not be the best way to deal
        # with multiple kinds
        return (self.datum.date(), self.datum.time(),
                self.getEditor(), self.getKindsAsText(),
                empty_if_none(h.timedelta_in_minutes(self.dauer)),
                F.html_to_text(empty_if_none(self.notiz)))

class Logbook(object):

    def __init__(self, mid=None):
        self.entries = []
        self.mid = mid
        self.masterId = None

    def getMasterId(self):
        return self.masterId

    def deleteEntryById(self, mid):
        db.execute(DELETE_LOGBOOK_ENTRY, {'mid': mid})

    def loadById(self, mid, sort_field='datum', sort_order='desc'):
        with db.cursor() as cursor:
            self.load_from_cursor(cursor, mid, sort_field, sort_order)

    def load_from_cursor(self, cursor, mid, sort_field='datum',
                         sort_order='desc'):
        self.mid = mid
        if (sort_field not in ALLOWED_SORTING
            or sort_order not in ALLOWED_SORTING):
            sort_field = 'datum'
            sort_order = 'desc'
        sql = LOAD_LOGBOOK % (sort_field, sort_order)
        cursor.execute(sql, (mid,))

        while True:
            row = cursor.fetchone()
            if not row:
                break
            entry = LogbookEntry()
            entry.setByRow(row)
            self.masterId = row[9]
            self.entries.append(entry)

    def hasEntries(self):
        return len(self.entries) > 0

    def getEntries(self):
        return self.entries

    def addEntry(self, entry):
        log.debug(entry)
        self.entries.append(entry)

    def store(self, uid):
        if self.mid is None:
            return
        with db.transaction() as cur:
            self.store_cursor(cur, uid)

    def store_cursor(self, cur, uid):
        for entry in self.entries:
            entry.store(cur, self.mid, uid)

    def getEvaluation(self):
        eval = LogbookEvaluation(self.mid)
        return eval.perform()

    def getTypEvaluation(self):
        eval = LogbookEvaluation(self.mid)
        return eval.perform_type_evaluation()


def build_logbook_entry(kinds, date, time, duration, notice, short_notice, typ):
    entry = LogbookEntry()
    entry.set_kinds(kinds)
    entry.setDate(datetime.combine(date, time))
    if duration is not None:
        entry.setDuration(duration)
    if notice is not None:
        entry.setNotice(notice)
    if short_notice is not None:
        entry.setShortNotice(short_notice)
    if typ is not None:
        entry.setTyp(typ)
    return entry


def text_subelement(parent, tag, text):
    """Create an element under parent with text content.
    The text will always be converted to unicode, so it should either be
    unicode or be an object where the default conversion to unicode is
    appropriate for XML serialisation.
    """
    if text is not None:
        element = ET.SubElement(parent, tag)
        element.text = unicode(text)

def iso_time_string(d):
    """Convert a datetime object to an ISO string."""
    # datetime.strftime() dislikes years before 1900, so we can't use it
    # here because the "unknown" date is before 1900.
    return ("%04d-%02d-%02dT%02d:%02d:%02d"
            % (d.year, d.month, d.day, d.hour, d.minute, d.second))

def string_iso_time(s):
    return datetime(*(time.strptime(s, "%Y-%m-%dT%H:%M:%S")[0:6]))


def export_logbook(case, case_id, cursor, selection, anonymize):
    """Add the contents of the case with id case_id to the elementtree case"""
    logbook = Logbook()
    logbook.load_from_cursor(cursor, case_id)

    logbook_element = ET.SubElement(case, "tagebuch")

    for entry in logbook.getEntries():
        entry_element = ET.SubElement(logbook_element, "tagebuch-eintrag")

        text_subelement(entry_element, "uuid_id", entry.uuid)
        text_subelement(entry_element, "tagebuch-datum",
                        iso_time_string(entry.datum))
        text_subelement(entry_element, "tagebuch-bearbeiter",
                        entry.bearbeiter_name)
        text_subelement(entry_element, "tagebuch-art",
                        ",".join(str(kind) for kind in entry.get_kinds()))
        text_subelement(entry_element, "tagebuch-sonstiges", entry.kurz_notiz)
        text_subelement(entry_element, "tagebuch-notiz", entry.notiz)
        if entry.dauer is not None:
            text_subelement(entry_element, "tagebuch-dauer",
                            int(h.timedelta_in_minutes(entry.dauer)))
        text_subelement(entry_element, "tagebuch-typ", entry.typ)

        export_logbook_documents(entry_element, entry.id, cursor, selection,
                                 anonymize)


def get_logbook_import_factory():
    """Return the component factory for the logbook for the libformed xmlimport
    """
    entry_node = StructureNode(new_node(RepeatNode, "tagebuch-eintrag"),
                               [new_node(TextLeaf, "tagebuch-datum"),
                                new_node(TextLeaf, "tagebuch-bearbeiter"),
                                new_node(TextLeaf, "tagebuch-art"),
                                new_node(TextLeaf, "tagebuch-sonstiges"),
                                new_node(TextLeaf, "tagebuch-notiz"),
                                new_node(IntLeaf, "tagebuch-dauer"),
                                new_node(IntLeaf, "tagebuch-typ"),
                                ],
                               [document_structure_node])
    return ComponentFactory(None,
                            StructureNode(None, [], [entry_node]),
                            StructureInstanceNode,
                            StructureInstanceTree)


xml_to_db = {
    "tagebuch-datum": "datum",
    "tagebuch-bearbeiter": "bearbeiter",
    "tagebuch-art": "art",
    "tagebuch-sonstiges": "kurz_notiz",
    "tagebuch-notiz": "notiz",
    "tagebuch-dauer": "dauer",
    "tagebuch-typ": "typ",
    }

def import_logbook(importer, case_tree, cursor, logbook_tree):
    """Import the logbook"""
    notiz_format = ("text/plain" if importer.cases_version < "6"
                    else "text/html")
    entries = logbook_tree.root.get_repeat_groups("tagebuch-eintrag")
    if entries:
        case_id = case_tree.getRootNode().getIdentifier()
        known = load_known_entries(case_id, cursor)
        for entry in entries.itervalues():
            if entry.uuid in known:
                entry_id = update_logbook_entry(case_id, cursor, entry,
                                                known[entry.uuid],
                                                notiz_format)
            else:
                entry_id = create_logbook_entry(case_id, cursor, entry,
                                                notiz_format)
            api = LogbookDocumentAPI(entry_id)
            import_documents_from_api(cursor, api, entry)


def load_known_entries(case_id, cursor):
    """Return a dictionary with all known logbook entries for the case case_id.
    The dictionary maps the uuids of the entries to their database ids.
    """
    cursor.execute("SELECT id, uuid_id FROM ka_logbuch_tbl_view"
                   " WHERE master_id = %(id)s", dict(id=case_id))
    known = dict()
    for id, uuid in cursor.fetchall():
        known[uuid] = id
    return known

def update_logbook_entry(case_id, cursor, entry, entry_id, notiz_format):
    """Update an existing logbook entry"""
    assignments = []
    values = dict(id=entry_id)
    for key, value in entry.leaves.iteritems():
        colname = xml_to_db.get(key)
        if colname is None:
            raise ValueError("Unknown logbook entry leaf: %r" % key)
        assignments.append("%s = %%(%s)s" % (colname, key))
        if value is not None:
            # We need to special case 'dauer' and 'datum' for now
            # because it cannot be handled by libformed alone yet.
            if colname == "dauer":
                value = timedelta(minutes=value)
            elif colname == "datum":
                value = string_iso_time(value)
            elif colname == "art":
                value = [int(item) for item in value.split(",")]
            elif colname == "notiz":
                if notiz_format == "text/plain":
                    converter = notiz_from_plaintext
                elif notiz_format == "text/html":
                    converter = clean_html
                else:
                    raise ValueError("unknown notiz_format %r"
                                     % (notiz_format,))
                value = converter(value)
        values[key] = value

    statement = ("UPDATE ka_logbuch_tbl_view SET %s WHERE id = %%(id)s"
                 % (", ".join(assignments),))
    cursor.execute(statement, values)

    return entry_id


def notiz_from_plaintext(plaintext):
    return "<p>" + unicode(F.nl2br(h.escape(plaintext))) + "</p>"

def create_logbook_entry(case_id, cursor, entry, notiz_format):
    """Create a new logbook entry."""
    cursor.execute("SELECT create_logbuch_ds(%(case_id)s, %(uuid)s)",
                   dict(case_id=case_id, uuid=entry.uuid))
    entry_id = cursor.fetchone()[0]
    return update_logbook_entry(case_id, cursor, entry, entry_id, notiz_format)


def export_as_xlsx(logbook, case):
    """Create a logbook xlsx export. Need case for meta data."""

    data = [["Tagebuch:", case.last_name + ", " + case.first_name]]
    data.append(["Datum:", date.today()])
    data.append([])

    # add columns header
    data.append(["Datum", "Uhrzeit", u"Bearbeiter/in", "Art",
                 "Dauer (min)", "Notizen"])

    for entry in logbook.getEntries():
        data.append(entry.get_data_for_xlsx())

    wb = Workbook()
    ws = wb.active
    ws.title = u"Tagebuch"

    alignment = Alignment(horizontal='general',
                          vertical='top',
                          text_rotation=0,
                          wrap_text=False,
                          shrink_to_fit=False,
                          indent=0)

    alignment_notice = Alignment(horizontal='general',
                          vertical='top',
                          text_rotation=0,
                          wrap_text=True,
                          shrink_to_fit=True,
                          indent=0)

    # Special column index
    DATE_COL=0
    TIME_COL=1
    DURATION_COL=4
    NOTICE_COL=5
    for row in range(len(data)):
        for col in range(len(data[row])):
            row_n = row + 1
            col_n = col + 1
            ws.cell(column=col_n, row=row_n).value = data[row][col]
            if col == NOTICE_COL:
                ws.cell(column=col_n, row=row_n).alignment = alignment_notice
            else:
                ws.cell(column=col_n, row=row_n).alignment = alignment

            if (col == 1 and row == 1) or col == DATE_COL:
                ws.cell(column=col_n, row=row_n).number_format = "DD.MM.YYYY"
            elif col == TIME_COL:
                ws.cell(column=col_n, row=row_n).number_format = "HH:MM"

        # Resize row with newlines in entry.notiz
        if row > 3:
            row_entry = ws.row_dimensions[row_n]
            linebreak_cnt = data[row][NOTICE_COL].count('\n')
            row_cnt_est   = len(data[row][NOTICE_COL])/92
            row_entry.height = 13 + (linebreak_cnt+row_cnt_est) * 11

    for col_letter in "ABE":
        column_entry = ws.column_dimensions[col_letter]
        column_entry.width = 12
    for col_letter in "CD":
        column_entry = ws.column_dimensions[col_letter]
        column_entry.width = 40
    column_entry = ws.column_dimensions['F']
    column_entry.width = 80

    return save_virtual_workbook(wb)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
