import unittest
import StringIO
import xml.etree.ElementTree as ET

from formed.meta.io.document import make_builder, read_document
from formed.instance.backends.xmlimport import XMLImporter

from mpulsweb.lib.export import Export, Table, TableContainer


class PlainExport(Export):

    def __init__(self, meta, tree, selection, one_table_per_page=False,
                 fill_enumerated_nodes=False, one_sheet_export=False,
                 squash_lines=False):
        self.tree = tree
        self.container = PlainContainer(
            meta, selection, Table,
            one_table_per_page=one_table_per_page,
            fill_enumerated_nodes=fill_enumerated_nodes,
            one_sheet_export=one_sheet_export,
            squash_lines=squash_lines)

    def export(self):
        self.container.fill(self.tree)
        return self.container


class PlainContainer(TableContainer):

    def use_master_id(self):
        return False

    def canonicalize(self):
        """Return the container contents in an easily compared representation.
        The return value is a list of triples, one for each table in the
        container, with each triple consisting of the table name, a
        tuple with the column names and a list of rows where each row is
        a list with one value for each column.
        """
        tables = []
        for name in self.table_order:
            if name in self.tables:
                table = self.tables[name]
                cols = tuple(col.name for col in table.columns)
                rows = []
                for row in table.rows:
                    rows.append([row.get(col) for col in cols])
                tables.append((name, cols, rows))
        return tables


class TabularExportTest(unittest.TestCase):

    formed_tree = """\
<?xml version="1.0" encoding="UTF-8"?>
<document>
  <case>
    <page description="1. Page A" name="page-1" containers="a,b,c,d">
      <choice description="1.1 Choice Field" name="choice1"
              function="suppress_value" size="1" target="a">
        <bool description="N/A" name="choice1-na" value="-1" checked="true"/>
        <bool description="Unknown" name="choice1-unknown" value="-3"
              invisible="true"/>
        <bool description="ABC" name="abc" value="1"/>
        <bool description="DEF" name="def" value="2"/>
      </choice>
      <text description="1.2 Some Text" name="text1" size="40"
            function="suppress_value" maxlength="60" target="b"/>
      <int description="1.3 Some Int" name="int1" function="suppress_value"
           target="c"/>
      <repeat description="1.3 Date Ranges" name="rg_date_ranges" target="d"
              containers="a,b">
        <date description="From" target="a" name="rg_dr_from"
              order=";rg_dr_till" function="suppress_value"/>
        <date description="Till" target="b" name="rg_dr_till"
              order=";rg_dr_from" function="suppress_value"/>
      </repeat>
    </page>
    <page description="2. Page B" name="page-2" containers="a,b,c">
      <float description="2.1 A Float" name="float2"
             function="suppress_value" target="a"/>
      <radio description="2.2 Radio" name="radio2" function="suppress_value"
             target="b" layout="horizontal">
        <bool description="Yes" name="yes" value="1"/>
        <bool description="No" name="no" value="0"/>
        <bool description="N/A" checked="true" name="na" value="-1"/>
      </radio>
      <repeat description="2.3 Outer" name="rg_outer" target="c"
              containers="a,b">
        <text description="2.3.1 Text" name="rg_outer_t" target="a"/>
        <repeat description="2.3.2 Inner" name="rg_inner" target="b"
                containers="a">
          <bool description="2.3.2.1 Bool" name="rg_inner_b" target="a"/>
        </repeat>
      </repeat>
    </page>
  </case>
  <logbook>
  </logbook>
</document>
"""

    def setUp(self):
        self.meta_tree = read_document(StringIO.StringIO(self.formed_tree),
                                       make_builder("", None),
                                       generate_rules=True,
                                       validate_tree=True)


class TabularExportTestBase(TabularExportTest):

    def setUp(self):
        super(TabularExportTestBase, self).setUp()
        self.data_tree = ET.fromstring(self.test_data)

    def check(self, expected, selection=None, one_table_per_page=False,
              fill_enumerated_nodes=False, one_sheet_export=False,
              squash_lines=False):
        if selection is None:
            selection = ["page-1", "page-2"]
        export = PlainExport(self.meta_tree, self.data_tree, selection,
                             one_table_per_page=one_table_per_page,
                             fill_enumerated_nodes=fill_enumerated_nodes,
                             one_sheet_export=one_sheet_export,
                             squash_lines=squash_lines)
        container = export.export()
        self.assertEqual(container.canonicalize(), expected)


class TestTabularExport(TabularExportTestBase):

    test_data = """\
<cases formed-type="http://mpuls.intevation.de/TestTabularExport" version="5">
  <case>
    <master>
      <choice1>-1</choice1>
      <text1>Lorem</text1>
      <int1>32</int1>
      <float2>1.25</float2>
      <radio2>0</radio2>
      <rg_date_ranges>
        <rg_dr_from>2015-01-01</rg_dr_from>
        <rg_dr_till>2015-02-28</rg_dr_till>
      </rg_date_ranges>
      <rg_outer>
        <rg_outer_t>egreditur</rg_outer_t>
        <rg_inner>
          <rg_inner_b>1</rg_inner_b>
        </rg_inner>
        <rg_inner>
          <rg_inner_b>0</rg_inner_b>
        </rg_inner>
        <rg_inner>
          <rg_inner_b>0</rg_inner_b>
        </rg_inner>
      </rg_outer>
      <rg_outer>
        <rg_outer_t>vultum</rg_outer_t>
        <rg_inner>
          <rg_inner_b>0</rg_inner_b>
        </rg_inner>
      </rg_outer>
    </master>
  </case>
</cases>
"""

    def test_allpages_logicaltables_multisheet_enumint(self):
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1', u'float2',
                      u'radio2'),
                     [[0, '-1', 'Lorem', '32', '1.25', '0']]),
                    (u'rg_date_ranges',
                     (u'rg_date_ranges_id', 'master_id', u'rg_dr_from',
                      u'rg_dr_till'),
                     [[0, 0, '2015-01-01', '2015-02-28']]),
                    (u'rg_outer',
                     (u'rg_outer_id', 'master_id', u'rg_outer_t'),
                     [[1, 0, 'egreditur'], [5, 0, 'vultum']]),
                    (u'rg_inner',
                     (u'rg_inner_id', u'rg_outer_id', u'rg_inner_b'),
                     [[2, 1, '1'],
                      [3, 1, '0'],
                      [4, 1, '0'],
                      [6, 5, '0']])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=False,
                   one_sheet_export=False)

    def test_page1_logicaltables_multisheet_enumint(self):
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1'),
                     [[0, '-1', 'Lorem', '32']]),
                    (u'rg_date_ranges',
                     (u'rg_date_ranges_id', 'master_id', u'rg_dr_from',
                      u'rg_dr_till'),
                     [[0, 0, '2015-01-01', '2015-02-28']])],
                   selection=["page-1"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=False,
                   one_sheet_export=False)

    def test_allpages_pagetables_multisheet_enumint(self):
        self.check([('page-1',
                     ('master_id', u'choice1', u'text1', u'int1'),
                     [[0, '-1', 'Lorem', '32']]),
                    (u'rg_date_ranges',
                     (u'rg_date_ranges_id', 'master_id', u'rg_dr_from',
                      u'rg_dr_till'),
                     [[0, 0, '2015-01-01', '2015-02-28']]),
                    ('page-2',
                     ('master_id', u'float2', u'radio2'),
                     [[0, '1.25', '0']]),
                    (u'rg_outer',
                     (u'rg_outer_id', 'master_id', u'rg_outer_t'),
                     [[1, 0, 'egreditur'], [5, 0, 'vultum']]),
                    (u'rg_inner',
                     (u'rg_inner_id', u'rg_outer_id', u'rg_inner_b'),
                     [[2, 1, '1'],
                      [3, 1, '0'],
                      [4, 1, '0'],
                      [6, 5, '0']])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=True,
                   fill_enumerated_nodes=False,
                   one_sheet_export=False)

    def test_allpages_logicaltables_singlesheet_enumint(self):
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1',
                      u'rg_dr_from', u'rg_dr_till',
                      u'float2', u'radio2',
                      u'rg_outer_t', u'rg_inner_b'),
                     [[0, '-1', 'Lorem', '32', None, None,
                       '1.25', '0', None, None],
                      [None, None, None, None, '2015-01-01', '2015-02-28',
                       None, None,None, None],
                      [None, None, None, None, None, None,
                       None, None, 'egreditur', None],
                      [None, None, None, None, None, None,
                       None, None, None, '1'],
                      [None, None, None, None, None, None,
                       None, None, None, '0'],
                      [None, None, None, None, None, None,
                       None, None, None, '0'],
                      [None, None, None, None, None, None,
                       None, None, 'vultum', None],
                      [None, None, None, None, None, None,
                       None, None, None, '0'],
                      ])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=False,
                   one_sheet_export=True)

    def test_allpages_logicaltables_singlesheet_enumtext(self):
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1',
                      u'rg_dr_from', u'rg_dr_till',
                      u'float2', u'radio2',
                      u'rg_outer_t', u'rg_inner_b'),
                     [[0, u'N/A', 'Lorem', '32', None, None,
                       '1.25', u'No', None, None],
                      [None, None, None, None, '2015-01-01', '2015-02-28',
                       None, None,None, None],
                      [None, None, None, None, None, None,
                       None, None, 'egreditur', None],
                      [None, None, None, None, None, None,
                       None, None, None, u'Ja'],
                      [None, None, None, None, None, None,
                       None, None, None, u'Nein'],
                      [None, None, None, None, None, None,
                       None, None, None, u'Nein'],
                      [None, None, None, None, None, None,
                       None, None, 'vultum', None],
                      [None, None, None, None, None, None,
                       None, None, None, u'Nein'],
                      ])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=True,
                   one_sheet_export=True)

    def test_allpages_logicaltables_singlesheet_enumint_squashed(self):
        self.maxDiff = None
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1',
                      u'rg_dr_from', u'rg_dr_till',
                      u'float2', u'radio2',
                      u'rg_outer_t', u'rg_inner_b'),
                     [[0, '-1', 'Lorem', '32', '2015-01-01', '2015-02-28',
                       '1.25', '0', 'egreditur', '1'],
                      [None, None, None, None, None, None,
                       None, None, None, '0'],
                      [None, None, None, None, None, None,
                       None, None, None, '0'],
                      [None, None, None, None, None, None,
                       None, None, 'vultum', '0'],
                      ])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=False,
                   one_sheet_export=True,
                   squash_lines=True)


class TestRepeatGroupSquash(TabularExportTestBase):

    test_data = """\
<cases formed-type="http://mpuls.intevation.de/TestTabularExport" version="5">
  <case>
    <master>
      <choice1>-1</choice1>
      <text1>Lorem</text1>
      <int1>32</int1>
      <float2>1.25</float2>
      <radio2>0</radio2>
      <rg_date_ranges>
        <rg_dr_from>2015-01-01</rg_dr_from>
        <rg_dr_till>2015-02-28</rg_dr_till>
      </rg_date_ranges>
      <rg_date_ranges>
        <rg_dr_from>2015-02-02</rg_dr_from>
      </rg_date_ranges>
      <rg_date_ranges>
        <rg_dr_till>2015-03-03</rg_dr_till>
      </rg_date_ranges>
      <rg_outer>
        <rg_outer_t>egreditur</rg_outer_t>
        <rg_inner>
          <rg_inner_b>1</rg_inner_b>
        </rg_inner>
        <rg_inner>
          <rg_inner_b>0</rg_inner_b>
        </rg_inner>
      </rg_outer>
    </master>
  </case>
</cases>
"""

    def test_repeat_group_squashing(self):
        '''Check that unrelated repeat groups are squashed together
        in more then one row.
        The important thing here is that the value of the second
        row of rg_inner_b is in the second row together with the
        value of rg_dr_from from the second rg_date_ranges group.

        Additionally this tests that the third rg_date_ranges
        value is not (wrongly) squashed into the second row.'''
        self.maxDiff = None
        self.check([('master',
                     ('master_id', u'choice1', u'text1', u'int1',
                      u'rg_dr_from', u'rg_dr_till',
                      u'float2', u'radio2',
                      u'rg_outer_t', u'rg_inner_b'),
                     [[0, '-1', 'Lorem', '32', '2015-01-01', '2015-02-28',
                       '1.25', '0', 'egreditur', '1'],
                      [None, None, None, None, '2015-02-02', None,
                       None, None, None, '0'],
                      [None, None, None, None, None, '2015-03-03',
                       None, None, None, None],
                      ])],
                   selection=["page-1", "page-2"],
                   one_table_per_page=False,
                   fill_enumerated_nodes=False,
                   one_sheet_export=True,
                   squash_lines=True)


if __name__ == "__main__":
    unittest.main()
