#!/usr/bin/env python

import sys
from optparse import OptionParser
import traceback
from datetime import timedelta
import decimal
import sqlite3 as db
import logging

import psycopg2 as dbapi
import psycopg2.extensions


log = logging.getLogger("import-sqlite")


dbapi.extensions.register_type(psycopg2.extensions.UNICODE)
dbapi.extensions.register_type(psycopg2.extensions.UNICODEARRAY)


# Add sqlite3 adapter for Decimal. Needed for float fields.
db.register_adapter(decimal.Decimal, float)


SELECT_MAX_ID = '''SELECT coalesce(max(id), -1) FROM %(relation)s'''

SELECT_FKZ = "SELECT id, database FROM fkz_tbl"

SELECT_STATUS = "SELECT status, zugriff FROM ka_status_tbl WHERE master_id = %(id)s"

INSERT_STATUS = '''
INSERT INTO ka_status_tbl (id, master_id, status, zugriff)
       VALUES (:id, :master_id, :status, :zugriff)'''

SELECT_LOGBOOK = '''
SELECT
    l.datum AS datum,
    md5(l.bearbeiter) AS bearbeiter,
    ARRAY(SELECT choice_id
            FROM nm_logbuch_art_tbl nm
           WHERE nm.parent_id = l.id) AS art,
    l.dauer AS dauer
FROM ka_logbuch_tbl l
WHERE l.master_id = %(master_id)s
'''

INSERT_LOGBOOK = '''
INSERT INTO ka_logbuch_tbl (id, master_id, datum, bearbeiter, dauer)
       VALUES (:id, :master_id, :datum, :bearbeiter, :dauer)
'''

INSERT_LOGBOOK_ART = '''
INSERT INTO nm_logbuch_art_tbl (parent_id, choice_id)
       VALUES (:parent_id, :choice_id)
'''


SELECT_NM_TAGS_CASES = '''
SELECT tag_id, created FROM nm_tags_cases WHERE master_id = %(id)s
'''

INSERT_NM_TAGS_CASES = '''
INSERT INTO nm_tags_cases (id, master_id, tag_id, created)
       VALUES (:id, :master_id, :tag_id, :created)
'''


ANON_TRANSFER = \
    "SELECT wert = 'on' FROM ka_konfiguration_tbl WHERE bez = 'anon_transfer'"

OUT_INSERT = "INSERT INTO %(tablename)s (%(fields)s) VALUES (%(placeholders)s)"

UPDATE_DATADUMP = "UPDATE dbinfo SET value = CURRENT_DATE WHERE key = 'datadump'"


BLACK_LIST = set([
    'id',
    'uuid_id',
    ])


# tables depending on master_tbl
DEFAULT_RELATIONS = [("master", None)]


# Relations that have a fkz column which should be set during export.
DEFAULT_FKZ_RELATIONS = set(["master"])


def table_name(relation):
    return relation + "_tbl"


def id_column(relation):
    return relation + "_id"


def get_max_id(cur, relation):
    stmt = SELECT_MAX_ID % {'relation': relation}
    cur.execute(stmt)
    row = cur.fetchone()
    if not row:
        return -1
    return row[0]


def usage(msg="", exit_code=0):
    print >> sys.stderr, """%s
usage:
%s -o <dump.db>
""" % (msg, sys.argv[0])
    sys.exit(exit_code)


def get_fkzs(cur):
    log.info("start reading fkzs")

    cur.execute(SELECT_FKZ)
    out = []
    while True:
        row = cur.fetchone()
        if not row:
            break
        out.append(row)

    log.info("finished reading fkzs")

    return out


def make_insert_statement(relation, blacklist, row_description):
    fields = []
    placeholders = []
    columns = []
    id_idx = None

    for idx, desc in enumerate(row_description):
        name = desc[0]
        if name == "id":
            id_idx = idx
            fields.append("id")
            placeholders.append(":id")
        if name in blacklist:
            continue
        columns.append((name, idx))
        placeholders.append(':%s' % name)
        fields.append(name)

    stmnt = OUT_INSERT % {"tablename": table_name(relation),
                          "fields": ','.join(fields),
                          "placeholders": ','.join(placeholders),
                          }

    return (stmnt, columns, id_idx)


def copy_relation(in_cur, out_cur, relation, parent, parent_mapping,
                  blacklist, extra_values=None, keep_fkz=False):
    log.info("start copying %r", relation)

    table = table_name(relation)
    out_counter = None
    stmnt = None
    mapping = {}

    parent_id_column = None
    if parent is not None:
        parent_id_column = id_column(parent)

    in_cur.execute("SELECT * FROM %(tablename)s_anonym_view"
                   % {"tablename": table})
    while True:
        row = in_cur.fetchone()
        if not row:
            break

        if stmnt is None:
            stmnt, columns, id_idx = \
                   make_insert_statement(relation, blacklist,
                                         in_cur.description)
            if id_idx is not None:
                out_counter = get_max_id(out_cur, table)

        values = {}
        if id_idx is not None:
            out_counter += 1
            mapping[row[id_idx]] = values["id"] = out_counter
        if extra_values is not None and keep_fkz:
            values.update(extra_values)
        for name, idx in columns:
            value = row[idx]
            if name == parent_id_column:
                value = parent_mapping[value]
            values[name] = value
        if extra_values is not None and not keep_fkz:
            values.update(extra_values)
        out_cur.execute(stmnt, values)


    log.info("finished copying %r", relation)

    return mapping


def copy_nm_tags_cases(in_cur, out_cur, cases):
    log.info("start copying nm_tags_cases")

    status_counter = get_max_id(out_cur, 'nm_tags_cases')

    for rid, nid in cases.items():
        in_cur.execute(SELECT_NM_TAGS_CASES, {'id': rid})
        for row in in_cur.fetchall():
            status_counter += 1
            tag_id, created = row
            out_cur.execute(INSERT_NM_TAGS_CASES, {
                'id': status_counter,
                'master_id': nid,
                'tag_id': tag_id,
                'created': created
                })
    log.info("finished copying nm_tags_cases")


def copy_status(in_cur, out_cur, cases):
    log.info("start copying ka_status_tbl")

    status_counter = get_max_id(out_cur, 'ka_status_tbl')

    for rid, nid in cases.items():
        in_cur.execute(SELECT_STATUS, {'id': rid})
        row = in_cur.fetchone()
        if not row:
            continue
        status_counter += 1
        status, zugriff = row
        out_cur.execute(INSERT_STATUS, {
            'id': status_counter,
            'master_id': nid,
            'status': status,
            'zugriff': zugriff
            })

    log.info("finished copying ka_status_tbl")


def to_seconds(td):
    if isinstance(td, timedelta):
        days = td.days * 86400
        secs = td.seconds
        ms = td.microseconds > 500000 and 1 or 0
        return days + secs + ms
    return td


def transfer_logbuch_tbl(in_cur, out_cur, cases):
    log.info("start copying ka_logbuch_tbl")

    new_lb_id = get_max_id(out_cur, 'ka_logbuch_tbl')

    for old_id, new_id in cases.items():
        in_cur.execute(SELECT_LOGBOOK, {'master_id': old_id})
        while True:
            row = in_cur.fetchone()
            if not row:
                break
            datum, bearbeiter, art, dauer = row
            new_lb_id += 1
            out_cur.execute(INSERT_LOGBOOK, {
                'id': new_lb_id,
                'master_id': new_id,
                'datum': datum,
                'bearbeiter': bearbeiter,
                'dauer': to_seconds(dauer)
                })
            for choice in art:
                out_cur.execute(INSERT_LOGBOOK_ART, {
                    'parent_id': new_lb_id,
                    'choice_id': choice,
                    })

    log.info("finished copying ka_logbuch_tbl")


def handle(out_cur, fkz, port, relations, blacklist, with_logs_tags, keep_fkz,
           fkz_relations):
    log.info("start copying DB %r, port %r", fkz, port)

    database = fkz[1]
    mappings = dict()

    in_con, in_cur = None, None
    try:
        in_con = dbapi.connect(database=database, port=port)
        in_cur = in_con.cursor()

        in_cur.execute(ANON_TRANSFER)
        row = in_cur.fetchone()
        if not row or not row[0]:
            log.info("No transfer allowed by %r", database)
            return

        # copy all depending relations
        for relation, parent in relations:
            if relation in fkz_relations:
                extra = {"fkz": fkz[0]}
            else:
                extra = None
            mappings[relation] = \
                  copy_relation(in_cur, out_cur, relation, parent,
                                mappings.get(parent), blacklist, extra,
                                keep_fkz)

        cases = mappings["master"]

        # copy ka_status_tbl
        copy_status(in_cur, out_cur, cases)

        if with_logs_tags:
            # copy logbook
            transfer_logbuch_tbl(in_cur, out_cur, cases)

            # copy tags
            copy_nm_tags_cases(in_cur, out_cur, cases)

    finally:
        if in_cur:
            in_cur.close()
        if in_con:
            in_con.close()

    log.info("finished copying DB %r, port %r", fkz, port)


DEFAULT_PORT = 5432


def parse_port_list(option, opt, value, parser):
    setattr(parser.values, option.dest, [int(p) for p in value.split(',')])


def parse_relation_list(option, opt, value, parser):
    relations = getattr(parser.values, option.dest)
    for item in value.split(","):
        parts = item.split(":")
        if len(parts) == 1:
            entry = (parts[0], "master")
        elif len(parts) == 2:
            entry = tuple(parts)
        else:
            raise SyntaxError("Too many colons in relation item")
        relations.append(entry)
    setattr(parser.values, option.dest, relations)


def parse_string_set(option, opt, value, parser):
    current_set = getattr(parser.values, option.dest)
    setattr(parser.values, option.dest, current_set | set(value.split(",")))


def main():
    parser = OptionParser()
    parser.set_defaults(port=[DEFAULT_PORT], relations=DEFAULT_RELATIONS,
                        blacklist=BLACK_LIST,
                        fkz_relations=DEFAULT_FKZ_RELATIONS,
                        log_level=logging.INFO)
    parser.add_option("-o", "--out",
                      help="Filename of the output database")
    parser.add_option("-p", "--port", action="callback",
                      callback=parse_port_list, type="string",
                      help=("Comma-separated port numbers of the PostgreSQL"
                            " databases"))
    parser.add_option("-r", "--relations", action="callback",
                      callback=parse_relation_list, type="string",
                      help=("Comma-separated list of additional relations"
                            " to export. Each relation may either be just the"
                            " relation name (with out the _tbl suffix) or"
                            " such a relation name, followed by a colon (:)"
                            " and the name of the parent relation"
                            " (also without _tbl suffix)."))
    parser.add_option("--blacklist", action="callback",
                      callback=parse_string_set, type="string",
                      help=("Comma-separated list of fields that should not"
                            " be exported."))
    parser.add_option("--no_logs_tags", action="store_false",
                      dest="logs_tags", default=True,
                      help="Don't export logs and tags tables")
    parser.add_option("--keep-fkz", action="store_true", default=False,
                      help=("Do not overwrite FKZ numbers read from the source"
                            " database."))
    parser.add_option("--fkz-relations", action="callback",
                      callback=parse_string_set, type="string",
                      help=("Comma-separated set of relations that have an"
                            " fkz column which should be set during export"
                            " in addition to master."))
    parser.add_option("--log-level", help="Logging level threshold.")

    opts, args = parser.parse_args()

    if opts.out is None:
        usage("Error: missing output database parameter -o", 1)

    logging.basicConfig(stream=sys.stderr, level=opts.log_level,
                        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")

    out_con, out_cur = None, None

    try:
        log.info("opening destination database %r", opts.out)
        out_con = db.connect(opts.out)
        out_cur = out_con.cursor()

        try:
            fkzs = get_fkzs(out_cur)
            for p in opts.port:
                for fkz in fkzs:
                    try:
                        handle(out_cur, fkz, p, opts.relations, opts.blacklist,
                               opts.logs_tags, opts.keep_fkz,
                               opts.fkz_relations)
                        log.exception("successfully exported data from DB %r,"
                                      " port %r", fkz, p)
                    except:
                        log.exception("failed to export data from DB %r, port %r",
                                      fkz, p)

            log.info("setting datadump timestamp")
            out_cur.execute(UPDATE_DATADUMP)

            log.info("start committing")
            out_con.commit()
            log.info("finished committing")
        except:
            log.info("rollback")
            out_con.rollback()
            raise
    finally:
        if out_cur:
            out_cur.close()
        if out_con:
            out_con.close()


if __name__ == '__main__':
    try:
        log.info("start export")
        try:
            main()
        except:
            log.exception("Exception during export")
            exit(1)
    finally:
        log.info("finished export")
